package EFS::CLI;

use strict;
use warnings;
use English;
use feature q{switch};

use MIME::Base64;

use Getopt::Long;
use IO::Pipe;
use Date::Format;
use Data::Dumper;
use List::Util qw(first shuffle);
use List::MoreUtils qw( uniq );
use Text::Abbrev;
use Text::FormatTable;
use POSIX qw(:errno_h);
use IO::File;
use IO::Socket::SSL;
use Socket;
use Storable qw(nfreeze);
use Readonly;
use Params::Validate qw( :all );
use Term::ReadPassword;
use File::stat;
use Config::Tiny;
use Try::Tiny;

use EFS::FQDN qw(fqdn);
use EFS::Logfile;
use EFS::Colors;
use EFS::Config::Client;
use EFS::Request;

use EFS::CLI::Any;
use EFS::CLI::AFSVolume;
use EFS::CLI::AutoRelease;
use EFS::CLI::Business;
use EFS::CLI::Cabinet;
use EFS::CLI::Campus;
use EFS::CLI::Cell;
use EFS::CLI::ChildRelease;
use EFS::CLI::Database;
use EFS::CLI::Dependency;
use EFS::CLI::Dependents;
use EFS::CLI::DistStats;
use EFS::CLI::Division;
use EFS::CLI::EFSServer;
use EFS::CLI::EFSServerCell;
use EFS::CLI::Feature;
use EFS::CLI::Fileqtree;
use EFS::CLI::Fileserver;
use EFS::CLI::Fileshare;
use EFS::CLI::ForbiddenFileRule;
use EFS::CLI::GroupRights;
use EFS::CLI::History;
use EFS::CLI::Home;
use EFS::CLI::Hostclass;
use EFS::CLI::Install;
use EFS::CLI::Location;
use EFS::CLI::Logfile;
use EFS::CLI::Metaproj;
use EFS::CLI::Namespace;
use EFS::CLI::Parameter;
use EFS::CLI::Platform;
use EFS::CLI::PlatformAlias;
use EFS::CLI::PlatformBuild;
use EFS::CLI::PlatformInstance;
use EFS::CLI::PlatformInstanceAlias;
use EFS::CLI::Pod;
use EFS::CLI::Project;
use EFS::CLI::PTS;
use EFS::CLI::Region;
use EFS::CLI::Release;
use EFS::CLI::ReleaseAlias;
use EFS::CLI::SSLServer;
use EFS::CLI::UserRights;

# NOTE: To be deprecated
use EFS::CLI::ReleaseLink;

#
# Globals
#
Readonly my $RETRY_MAX           => 10;
Readonly my $RETRY_SLEEP         => 6;
Readonly my $TIMEOUT             => 60;
Readonly my $SPACE               => q{ };
Readonly my $MAX_PASSWORD_RETRY  => 3;

my $hostname = fqdn;

my @commands_common =
    qw( display setattr getattr delattr history );
my @commands_define =
    ( qw( define undef ),   @commands_common );
my @commands_create =
    ( qw( create destroy ), @commands_common );
my @commands_move   =
    qw( move moveall set_movelock clear_movelock );

my @commands_notify =
    qw( set_owner_list set_notify_list test_owner_list
        get_owner_list get_notify_list test_notify_list );

my @commands_clear_notify =
    qw( clear_owner_list clear_notify_list);

my @commands_mpr =
    qw( chown disttree diststate execlinks audit purge list
        set_noworldaccess set_distgroup clear_distgroup );

# Commands that don't require authorization
our @unauthorized_commands =
    qw( display getattr history showtree diststate list status
        get_owner_list get_notify_list );

# Must be our, not my...   efsd reads this hash now to validate method calls
our %commands = (

    any => { map { $_ => 1 } (
        qw( history diststate dist execlinks
            audit purge list disttree )
    ) },

    # XXX: We need some commands for displaying AFS volumes...
    afsvolume => { },

    autorelease => { map { $_ => 1 } (
        qw( create dist revert )
    ) },

    business => { map { $_ => 1 } @commands_define },

    cabinet => { map { $_ => 1 } @commands_define },

    cell => { map { $_ => 1 } (
        @commands_define,
        qw( dumpconfig getnames execlinks init uninit
            expose hide diffcellstate clonecellstate )
    ) },

    childrelease => { create => 1 },

    campus => { map { $_ => 1 } @commands_define },

    database => { map { $_ => 1 } qw( backup dumpconfig ) },

    dependents => { map { $_ => 1 } qw( display showtree ) },

    dependency => { map { $_ => 1 } (
        qw( define batchdefine undef
            diststate batchundef display
            showtree audit auditcircular )
    ) },

    diststats => { map { $_ => 1 } qw( display ) },

    division => { map { $_ => 1 } @commands_define },

    efsserver => { map { $_ => 1 } (
        @commands_define,
        qw( dumpconfig restart checkaccess status initssl archivelogs newkeys
            purgekeys newcerts purgecerts newdbpw showcreds dumpmaps )
    ) },

    efsservercell => { map { $_ => 1 } @commands_define },

    feature => { map { $_ => 1 } @commands_define },

    ffrule => { map { $_ => 1 } @commands_define },

    fileqtree => { map { $_ => 1 } (
        @commands_define, qw( init execlinks )
    ) },

    fileserver => { map { $_ => 1 } (
        @commands_define, qw( showmount execlinks )
    ) },

    fileshare => { map { $_ => 1 } (
        @commands_define, qw( showexports execlinks )
    ) },

    grouprights => { map { $_ => 1 } @commands_define },

    history => { map { $_ => 1 } qw( audit correct ) },

    home => { map { $_ => 1 } (
        @commands_create, @commands_move,
        qw( recover purge ),
    ) },

    hostclass => { map { $_ => 1 } @commands_define },

    install => { map { $_ => 1 } (
        @commands_create, qw( chown audit unclone )
    ) },

    location => { map { $_ => 1 } @commands_define },

    logfile => { map { $_ => 1 } qw( history ) },

    metaproj => { map { $_ => 1 } (
        @commands_create, @commands_notify, @commands_mpr,
        qw( clear_notify_list define undef ),
    ) },

    namespace => { map { $_ => 1 } qw( rebuild ) },

    parameter => { map { $_ => 1 } @commands_define },

    platform => { map { $_ => 1 } (
        @commands_define, qw( getnames rename )
    ) },

    platformbuild => { map { $_ => 1 } @commands_define },

    platformalias => { map { $_ => 1 } @commands_define },

    platforminstance => { map { $_ => 1 } (
        @commands_define, qw( getnames rename )
    ) },

    platforminstancealias => { map { $_ => 1 } @commands_define },

    pod => { map { $_ => 1 } @commands_define },

    project => { map { $_ => 1 } (
        @commands_create, @commands_notify,
        @commands_clear_notify, @commands_mpr,
        @commands_move, qw( set_dist2dev ),
    ) },

    pts => { map { $_ => 1 } (
        # EFS specific stuff
        qw( ),
        # Standard PTS methods
        qw( adduser apropos chown creategroup createuser delete examine help
            interactive listentries listmax listowned membership quit removeuser
            rename setfields setmax sleep source )
    ) },

    region => { map { $_ => 1 } @commands_define },

    release => { map { $_ => 1 } (
        @commands_create, @commands_mpr,
        qw( dist checkpoint recover upload download
            lock unlock setstage clone unclone )
    ) },

    releasealias => { map { $_ => 1 } (
        @commands_create, qw( dist disttree diststate )
    ) },

    # NOTE: To be deprecated
    releaselink => { map { $_ => 1 } (
        @commands_create, qw( dist disttree diststate )
    ) },

    sslserver => { map { $_ => 1 } (
        @commands_define, qw( initca showcerts )
    ) },

    userrights => { map { $_ => 1 } @commands_define },

);

my %CLI_ORDER = (

    autorelease => {
        default => [qw( metaproj project )],
        revert  => [qw( metaproj project name )],
    },

    business => {
        default => [qw( name )],
        define  => [qw( name description )],
    },

    cabinet => {
        default => [qw( location name )],
    },

    campus => {
        default => [qw( region name )],
        define  => [qw( region name description )],
    },

    cell => {
        default        => [qw( name )],
        clonecellstate => [qw( fromcell tocell )],
        define         => [qw( region campus location name
                               celltype fstype parentcell ) ],
    },

    childrelease => {
        create => [qw( metaproj project name parent )],
    },

    dependents => {
        default => [qw( metaproj project release type )],
    },

    dependency => {
        default => [
            qw( dependent_metaproj dependent_project
                dependent_release metaproj project
                release type )
        ],
        batchdefine => [
            qw( dependent_metaproj dependent_project
                dependent_release type dependencies )
        ],
        undef => [
            qw( dependent_metaproj dependent_project
                dependent_release metaproj project
                release type )
        ],
        batchundef => [
            qw( dependent_metaproj dependent_project
                dependent_release type dependencies )
        ],
        diststate => [qw( metaproj project release )],
        showtree  => [qw( metaproj project release type )],
    },

    diststats => {
        display => [qw( history )]
    },

    division => {
        default => [qw( business name )],
        define  => [qw( business name description )],
    },

    efsserver => {
        default    => [qw( name )],
        define     => [qw( name hosttype )],
        dumpmaps   => [qw( name hosttype )],
        dumpconfig => [],
    },

    efsservercell => {
        default => [qw( efsserver cell )],
    },

    feature => {
        default => [qw( name )],
        define  => [qw( name value )],
    },

    ffrule => {
        default => [qw( name )],
        define  => [qw( name regex description )],
    },

    fileqtree => {
        default => [qw( cell fileserver fileshare name )],
    },

    fileserver => {
        default => [qw( cell name )],
        define  => [qw( cell name )],
    },

    fileshare => {
        default => [qw( cell fileserver name )],
        define  => [qw( cell fileserver name filetype )],
    },

    grouprights => {
        default => [qw( metaproj role groupname )],
    },

    home => {
        default => [qw( name )],
        create  => [qw( name cell )],
        move    => [qw( name cell )],
        moveall => [qw( fromcell tocell )],
    },

    hostclass => {
        default => [qw( name )],
        define  => [qw( name description )],
    },

    install => {
        default => [qw( metaproj project release name )],
    },

    location => {
        default => [qw( region campus name )],
        define  => [qw( region campus name description )],
    },

    logfile => {
        default => [qw( id )],
    },

    metaproj => {
        default           => [qw( name )],
        create            => [qw( name username groupname owner_list )],
        set_owner_list    => [qw( name address )],
        set_notify_list   => [qw( name address )],
        test_owner_list   => [qw( name )],
        test_notify_list  => [qw( name )],
        set_distgroup     => [qw( name distgroup )],
        set_noworldaccess => [qw( name noworldaccess )],
        # define has been deprecated
        define            => [qw( name username groupname owner_list )],
    },

    namespace => {
        default => [qw( type )],
    },

    parameter => {
        default => [qw( name )],
        define  => [qw( name value )],
    },

    platform => {
        default => [qw( name )],
        rename  => [qw( name newname )],
        define  => [qw( name backwards emulates )],
    },

    platformalias => {
        default => [qw( name )],
        define  => [qw( name platform )],
    },

    platformbuild => {
        default => [qw( name )],
        define  => [qw( name platform kernel initrd options )],
    },

    platforminstance => {
        default => [qw( name )],
        rename  => [qw( name newname )],
        define  => [qw( name platform backwards emulates )],
    },

    platforminstancealias => {
        default => [qw( name )],
        define  => [qw( name platforminstance )],
    },

    pod => {
        default => [qw( region campus location name )],
        define  => [qw( region campus location name description )],
    },

    project => {
        default           => [qw( metaproj name )],
        create            => [qw( metaproj name cell )],
        move              => [qw( metaproj name cell )],
        moveall           => [qw( fromcell tocell )],
        set_owner_list    => [qw( metaproj name address )],
        set_notify_list   => [qw( metaproj name address )],
        test_owner_list   => [qw( metaproj name )],
        test_notify_list  => [qw( metaproj name )],
        set_movelock      => [qw( metaproj name efsserver pid fromcell tocell )],
        set_distgroup     => [qw( metaproj name distgroup )],
        set_noworldaccess => [qw( metaproj name noworldaccess )],
        set_dist2dev      => [qw( metaproj name dist2dev )],
    },

    region => {
        default => [qw( name )],
        define  => [qw( name description )],
    },

    releasealias => {
        default => [qw( metaproj project name )],
        create  => [qw( metaproj project name target )],
    },

    # NOTE: To be deprecated
    releaselink => {
        default => [qw( metaproj project name )],
        create  => [qw( metaproj project name target )],
    },

    release => {
        default           => [qw( metaproj project name )],
        clone             => [qw( metaproj project name clone )],
        setstage          => [qw( metaproj project name stage )],
        set_distgroup     => [qw( metaproj project name distgroup )],
        set_noworldaccess => [qw( metaproj project name noworldaccess )],
    },

    sslserver => {
        default => [qw( type )],
        define  => [qw( type efsserver )],
    },

    userrights => {
        default => [qw( metaproj role username )],
    },

);

my @common_dist_args =
    qw( region=s campus=s location=s cells|cell=s
        celltype=s hiddencells global noefsops dryrun );

my @common_diststate_args =
    ( @common_dist_args, qw( details=s verbose|v debug ) );

my @common_dumpconfig_args =
    ( @common_dist_args, qw( force nodist ), );

my @common_list_args = qw( devpaths showlinks all );

my %CLI_OPTIONS = (

    autorelease => {
        revert => [qw( nodestroy force )],
    },

    any => {
        dist      => [@common_dist_args],
        disttree  => [@common_dist_args],
        diststate => [@common_diststate_args],
        execlinks => [qw( force )],
        audit     => [qw( audits=s force verbose|v )],
        list      => [@common_list_args],
    },

    cell => {
        display        => [@common_dist_args],
        diffcellstate  => [@common_dist_args],
        clonecellstate => [qw( force )],
        execlinks      => [qw( force )],
        getnames       => [@common_dist_args],
    },

    database => {
        dumpconfig => [@common_dumpconfig_args, qw( file=s )],
    },

    dependency => {
        display     => [qw( direct )],
        batchdefine => [qw( complete )],
        diststate   => [@common_diststate_args],
        showtree    => [ @common_dist_args, qw( maxdepth=i conflicts=s ) ],
    },

    dependents => {
        display  => [qw( direct )],
        showtree => [qw( maxdepth=i )],
    },

    diststats => {
        display => [qw( detail=s sortby=s )],
    },

    efsserver => {
        dumpconfig => [@common_dumpconfig_args],
        restart    => [qw( hosttype=s update )],
    },

    fileserver => {
        showmount => [qw( verbose|v celltype=s )],
        execlinks => [qw( force )],
    },

    fileshare => {
        execlinks => [qw( force )],
    },

    fileqtree => {
        execlinks => [qw( force )],
    },

    grouprights => {
        default => [qw( noefsops )],
    },

    groupname => {
        display => [qw( verbose conflicts )],
    },

    history => {
        default => [
            qw( starttime=s endtime=s startdate=s enddate=s
                byuser=s bygroup=s efshost=s efsversion=s
                request=s command=s status=i
                count=i orderby=s verbose|v all days=i
                weeks=i failed report=s parent=i )
        ],
    },

    home => {
        purge => [qw( now )],
    },

    install => {
        create     => [qw( force )],
        display    => [qw( sortby=s )],
        lock       => [qw( verbose|v )],
        unlock     => [qw( verbose|v username=s )],
        chown      => [qw( verbose|v username=s )],
        checkpoint => [qw( nolock noaudit )],
        audit      => [qw( audits=s force fakereadelf=s )],
    },

    location => {
        dumpconfig => [@common_dumpconfig_args],
    },

    metaproj => {
        audit     => [qw( audits=s force verbose|v )],
        define    => [qw( notify_list )],
        display   => [qw( sortby=s )],
        chown     => [qw( recurse verbose|v )],
        dist      => [@common_dist_args],
        disttree  => [@common_dist_args],
        diststate => [@common_diststate_args],
        execlinks => [@common_diststate_args, qw( force )],
        list      => [@common_list_args],
    },

    namespace => {
        rebuild => [@common_diststate_args],
    },

    pod => {
        display => [qw( name=s )],
        history => [qw( name=s )],
    },

    project => {
        display   => [qw( sortby=s )],
        destroy   => [qw( verbose|v )],
        chown     => [qw( norecurse verbose|v )],
        audit     => [qw( audits=s force verbose|v )],
        dist      => [@common_dist_args],
        disttree  => [@common_dist_args],
        diststate => [@common_diststate_args],
        execlinks => [@common_diststate_args, qw( force )],
        list      => [@common_list_args],
        move      => [qw( nobuild )],
    },

    releasealias => {
        dist      => [ @common_dist_args,
                      qw( no-strict-releasealiases|nsra|nsrl ) ],
        disttree  => [@common_dist_args],
        diststate => [@common_diststate_args],
    },

    # NOTE: to be deprecated
    releaselink => {
        dist      => [ @common_dist_args,
                       qw( no-strict-releaselinks|nsrl ) ],
        disttree  => [@common_dist_args],
        diststate => [@common_diststate_args],
    },

    release => {
        audit      => [qw( audits=s force verbose|v fakereadelf=s )],
        checkpoint => [qw( query nolock noaudit )],
        chown      => [qw( norecurse verbose|v username=s )],
        clone      => [qw( src build force )],
        create     => [qw( clonesrc=s )],
        destroy    => [qw( force )],
        display    => [qw( sortby=s )],
        dist       => [
            @common_dist_args,
            qw( verbose|v no-strict-releasealiases|nsra|nsrl noclone )
        ],
        disttree  => [@common_dist_args],
        diststate => [@common_diststate_args],
        execlinks => [@common_diststate_args, qw( force )],
        list      => [@common_list_args],
        lock      => [qw( norecurse verbose|v )],
        purge     => [qw( now noefsops )],
        unlock    => [qw( norecurse verbose|v username=s )],
        upload    => [qw( force recurse packageonly )],
        download  => [qw( force norecurse )],
    },

    sslserver => {
        initca => [qw( force )],
    },

    userrights => {
        default => [qw( noefsops )],
    },

    username => {
        display => [qw( verbose conflicts metaowners )],
    },

);

# Common validation spec for ALL methods

our %validate_common = (
    noefsops   => { type => SCALAR, optional => 1 },
    history    => { type => SCALAR, optional => 1 },
    authuser   => { type => SCALAR, optional => 1 },
    authorized => { type => SCALAR, optional => 1 },
);

# Common validation spec for the dist CLI agrs

our %validate_dist_args_common = (
    cells       => { type => SCALAR | ARRAYREF | UNDEF, optional => 1 },
    region      => { type => SCALAR,                    optional => 1 },
    campus      => { type => SCALAR,                    optional => 1 },
    location    => { type => SCALAR,                    optional => 1 },
    global      => { type => BOOLEAN,                   optional => 1 },
    hiddencells => { type => BOOLEAN,                   optional => 1 },
);

# Factor all the -celltype stuff together

##  no critic: FixedStringMatches
Readonly my $CELLTYPE_REGEX => qr{\A(?: all|dev|prod|test )\z}msx;

Readonly my @CELLTYPE_ARG_DEFAULT_NONE => (
    type     => SCALAR | UNDEF,
    regex    => $CELLTYPE_REGEX,
    optional => 1,
);

Readonly my @CELLTYPE_ARG_DEFAULT_ALL =>
    ( @CELLTYPE_ARG_DEFAULT_NONE, default => q{all}, );

Readonly my @CELLTYPE_ARG_DEFAULT_DEV =>
    ( @CELLTYPE_ARG_DEFAULT_NONE, default => q{dev}, );

our %validate_download_args = (
    celltype => {@CELLTYPE_ARG_DEFAULT_NONE},
    %validate_dist_args_common,
);

our %validate_diststate_args = (
    celltype => {@CELLTYPE_ARG_DEFAULT_ALL},
    %validate_dist_args_common,
);

our %validate_dist_args = (
    dryrun   => { type => SCALAR, optional => 1 },
    celltype => {@CELLTYPE_ARG_DEFAULT_DEV},
    %validate_dist_args_common,
);

our %validate_purge_args = (
    noefsops => { type => SCALAR, optional => 1 },
    celltype => {@CELLTYPE_ARG_DEFAULT_ALL},
    %validate_dist_args_common,
);

our %validate_dumpconfig_args = (
    force    => { type => BOOLEAN, optional => 1 },
    nodist   => { type => BOOLEAN, optional => 1 },
    celltype => {@CELLTYPE_ARG_DEFAULT_ALL},
    %validate_dist_args_common,
);

my @CLIENT_OPTIONS =
    qw( help serverhost=s serverport=s servertype=s servercell=s serverauth=s
        authuser=s no-auth-cache );

sub get_options_plan {

    my $class  = shift;
    my $type   = shift;
    my $action = shift || q{default};

    my @cli_order = $class->cli_order(
        type   => $type,
        action => $action,
    );

    my @plan = map { qq{$_=s} } @cli_order;

    if ( exists $CLI_OPTIONS{$type} ) {
        if ( exists $CLI_OPTIONS{$type}->{$action} ) {
            push ( @plan, @{ $CLI_OPTIONS{$type}->{$action} } );
        }
        elsif ( exists $CLI_OPTIONS{$type}->{default} ) {
            push ( @plan, @{ $CLI_OPTIONS{$type}->{default} } );
        }
    }

    if ( $action eq q{history} ) {
        push ( @plan, @{ $CLI_OPTIONS{history}->{default} } );
    }

    return uniq sort @plan;

}

sub cli_order {

    my $class = shift;
    my %args  = @_;

    my $attribute = $args{type} eq q{attribute} ? 1 : 0;

    my $type = $attribute ? $args{parent_type} : $args{type};
    my $action = $args{action} || q{default};

    my @cli_order = ();

    my $cli_order = $CLI_ORDER{$type} || {};

    if ( $attribute || $action =~ m/\A (set|get|del) attr \z/msx ) {
        foreach my $key ( @{ $cli_order->{default} } ) {
            if ( $key eq q{name} ) {
                $key = $type;
            }
            # We just need to make sure the default entry has the
            # primary keys only
            push @cli_order, $key;
        }
        push @cli_order, qw(name value);
    }
    elsif ( defined $cli_order->{$action} ) {
        @cli_order = @{ $cli_order->{$action} };
    }
    elsif ( defined $cli_order->{default} ) {
        @cli_order = @{ $cli_order->{default} };
    }

    return @cli_order;

}

sub new {

    my $class = shift;

    my $config = EFS::Config::Client->new;

    # XXX: Help processing, which is a huge mess
    if ( scalar @ARGV == 0 ) {

        # efs
        do_help({
            mode        => q{efs},
            server_args => undef,
            argv        => \@ARGV,
            action      => undef,
            object_type => undef
        });
        exit 0;
    }

    if (    ( scalar @ARGV == 1 )
        and ( defined $ARGV[0] and $ARGV[0] eq q{help} ) ) {

        # efs help
        do_help({
            mode        => q{efs help},
            server_args => undef,
            argv        => \@ARGV,
            action      => undef,
            object_type => undef
        });
        exit 0;
    }

    if (    ( scalar @ARGV == 2 )
        and ( defined $ARGV[0] )
        and ( $ARGV[0] eq q{help} )
        and ( defined $ARGV[1] ) ) {

        # efs help object
        if ( defined $commands{ $ARGV[1] } ) {
            do_help({
                mode        => q{efs help object},
                server_args => undef,
                argv        => \@ARGV,
                action      => undef,
                object_type => $ARGV[1]
            });
        }
        else {
            die qq{$ARGV[1] is not a valid object\n};
        }
        exit 0;
    }


    my %args = ( servercell => $config->get_value( q{client}, q{cell} ) );
    my %server_args = (
        serverauth => $config->get_value( q{security}, q{authtype} ) || q{GSSAPI},
        serverport => get_serverport(),
    );

    my $request       = join ( q{ }, @ARGV );
    my ($action)      = shift (@ARGV);
    my ($object_type) = shift (@ARGV);

    my @plan      = $class->get_options_plan( $object_type, $action );
    my @cli_order = $class->cli_order(
        type   => $object_type,
        action => $action,
    );

    GetOptions( \%args, @plan, @CLIENT_OPTIONS )
        || die qq{Unable to parse \@ARGV\n};

    foreach my $key (qw( serverhost servertype servercell
                         serverport serverauth authuser no-auth-cache )) {

        if ( $args{$key} ) {
            $server_args{$key} = delete $args{$key};
        }
        elsif ( $key eq q{no-auth-cache} ) {

            my $env_key = q{EFS_} . uc ($key);
            $env_key =~ s{-}{_}gmsx;

            if ( $ENV{$env_key} ) {
                $server_args{$key} = $ENV{$env_key};
            }

        }
        elsif ( $key =~ m{^server}msx ) {

            my $env_key = q{EFS_} . uc ($key);

            if ( $ENV{$env_key} ) {
                if ( not $ENV{EFS_TEST_SUITE} ) {
                    warn(
                        qq{Environment variable $env_key ($ENV{$env_key}) },
                        qq{overrides default $key.\n}
                    );
                }
                $server_args{$key} = $ENV{$env_key};
            }

        }

    }

    if ( not $server_args{serverauth} ~~ [qw( GSSAPI PAM )] ) {
        die qq{Unsupported serverauth '$server_args{serverauth}'\n};
    }

    if ( defined $args{help} ) {
        do_help({
            mode        => q{efs action object --help},
            server_args => \%args,
            argv        => \@ARGV,
            action      => $action,
            object_type => $object_type
        });
        exit 0;
    }

    while (@ARGV) {

        my ($value) = shift (@ARGV);
        my $key;

        if (@cli_order) {
            $key = shift (@cli_order);
            if ( exists $args{$key} ) {
                # Given by a -flag already
                die qq{Extraneous command line argument: '$value'\n};
            }
        }
        else {
            die qq{Extraneous command line argument: '$value'\n};
        }

        $args{$key} = $value;

    }

    # NOTE: We delete all these %args entries, since we do NOT want them
    # passes to the server as part of the command args

    my $redirected       = 0;
    my $invalid_password = 0;

    my $try_gssapi       = $server_args{serverauth} =~ m{\bGSSAPI\b}ms; 

CONNECTION: {

        if ( $redirected && $ENV{EFS_CLIENT_DEBUG} ) {
            STDOUT->print(
                qq{[DEBUG] Reconnecting to servercell $server_args{servercell}\n}
            );
        }

        my @servers = ();

        if ( $server_args{serverhost} ) {
            @servers = ( $server_args{serverhost} );
        }
        else {
            @servers = get_servernames(
                servercell => $server_args{servercell},
                servertype => $server_args{servertype},
            );
        }

        if ( not @servers ) {
            die qq{Unable to determine which servers to connect to.\n};
        }

        my $connection;
        my $efsserver;

      SERVER: foreach my $server (@servers) {

            my $retry = 0;

          PORT: {

                $connection = IO::Socket::SSL->new(
                    PeerAddr    => $server,
                    PeerPort    => $server_args{serverport},
                    Proto       => q{tcp},
                    Type        => SOCK_STREAM,
                    Timeout     => $TIMEOUT,
                    SSL_version => q{SSLv3},
                );

                if ($connection) {
                    if ( $ENV{EFS_CLIENT_DEBUG} ) {
                        STDOUT->print(
                            q{[DEBUG] Connected to },
                            qq{$server:$server_args{serverport}\n} );
                    }
                    $efsserver = $server;
                    last SERVER;
                } else {
                    if (    ( $retry < $RETRY_MAX )
                                and ( $ERRNO == ECONNREFUSED or $ERRNO == EINTR ) ) {
                        STDOUT->print(
                            q{Connection refused by port },
                            qq{$server_args{serverport} },
                            qq{on server $server (retrying...)\n}
                        );
                        sleep ($RETRY_SLEEP);
                        $retry++;
                        redo PORT;
                    } else {
                        STDOUT->print(
                            q{Unable to connect to efsd on server: },
                            qq{$server, port: $server_args{serverport}: $EVAL_ERROR\n}
                        );
                        next SERVER;
                    }
                }
            }

        }                       # foreach my $server ( @servers ) {

        if ( not $connection ) {
            die(
                q{Unable to connect to efsd on any },
                qq{of the following servers:\n\t},
                join ( qq{\n\t}, @servers ), qq{\n}
            );
        }

        $connection->sockopt( SO_KEEPALIVE, 1 ) || die qq{Unable to set keepalive: $ERRNO\n};

        $SIG{INT}  = \&sigint_handler;
        $SIG{PIPE} = q{IGNORE};

        $connection->print( qq{CLI\n} );

        if ( $try_gssapi ) {

            $connection->print( qq{GSSAPI\n} );

            my $eval_error;
            try {
                EFS::Request->authenticate_gssapi_client($connection,$efsserver);
            }
            catch {
                $eval_error = $_;
            };

            if ( $eval_error ) {
                warn( qq{GSSAPI Authentication Failure:\n}, $eval_error );
                $try_gssapi = 0;
                redo CONNECTION;
            }

        }
        else {

            $connection->print( qq{PAM\n} );

            my ( $authuser, $authpass );

            if ( $server_args{authuser} ) {
                $authuser = $server_args{authuser};
                $authpass = $class->get_authuser_password(
                    authuser         => $authuser,
                    invalid_password => $invalid_password,
                );
            } else {
                ($authuser,$authpass) = $class->get_username_password(
                    invalid_password => $invalid_password,
                    q{no-auth-cache} => $server_args{ q{no-auth-cache} },
                );
            }

            if ( not $authuser or not $authpass ) {
                die qq{Unable to determine username and password for CLI/PAM authentication\n};
            }

            $connection->print( qq{$authuser\n} );
            $connection->print( qq{$authpass\n} );

            if ( $ENV{EFS_TEST_FAKE_PASSWORD} ) {
                die qq{Access denied\n};
            }

        }

        $connection->autoflush(1);
        STDERR->autoflush(1);
        STDOUT->autoflush(1);

        my %client_data = (

            metadata => {
                envvars  => get_envvars(),
                cell     => $config->get_value( q{client}, q{cell} ),
            },

            operation => {
                object_type => $object_type,
                request     => $request,
                action      => $action,
                args        => \%args,
            },

        );

        my $frozen_data = nfreeze( \%client_data );
        $frozen_data = pack ( q{N/a*}, $frozen_data );
        $connection->print($frozen_data);

        my $seen_exit_status = undef;

      GETLINE: while ( defined ( my $line = $connection->getline ) ) {

            given ( $line ) {

                when ( m{^ _EFS_WAITFOR: \h+ (\S+)}msx ) {
                    if ( not exists $ENV{EFS_CLIENT_NO_WAITDIR} ) {
                        EFS::Util->WaitDir( directory => $1 );
                    }
                }

                when ( m{^ _EFS_EXIT_STATUS: \s+ (\d+)}x ) {
                    if ( $ENV{EFS_CLIENT_DEBUG} ) {
                        STDOUT->print( qq{[DEBUG] Received status message: $_\n} );
                    }
                    $seen_exit_status = $1;
                    last GETLINE;
                }

                when ( m{^ _EFS_INVALID_PASSWORD:}x ) {
                    $invalid_password++;
                    $connection->close;
                    if ( $invalid_password > $MAX_PASSWORD_RETRY ) {
                        die qq{Access denied\n};
                    }
                    redo CONNECTION;
                }

                when ( m{^ _EFS_REDIRECT_(CELL|SERVER): \s+ ([\w\.]+)}x ) {

                    my ($type,$target) = ($1,$2);
                    my $key = $type eq q{CELL} ? q{servercell} : q{serverhost};

                    if ( $ENV{EFS_CLIENT_DEBUG} ) {
                        STDOUT->print( qq{[DEBUG] Received redirect to $type $target\n} );
                    }

                    $redirected and die(
                        qq{We've already been redirected once.\n},
                        q{Multiple redirection is not supported. },
                        qq{Contact EFS Operations\n},
                    );

                    $redirected++;
                    $server_args{$key} = $target;
                    $connection->close;

                    redo CONNECTION;

                }

                default {
                    STDOUT->print( $_ );
                }

            }

        }

        if ( defined $seen_exit_status ) {
            exit $seen_exit_status;
        }
        else {
            die qq{Exit status from efs server not received. Assuming failure...\n};
        }

    }

}

sub do_help {

    my $params = shift;

    my $mode        = $params->{mode} || q{efs};
    my $server_args = $params->{server_args};
    my $argv        = $params->{argv};
    my $action      = $params->{action};
    my $object_type = $params->{object_type};
    my $cli_order   = $params->{cli_order};

    if ( ( $mode eq q{efs} ) or ( $mode eq q{efs help} ) ) {

        print qq{\n};
        foreach my $object ( sort keys %commands ) {
            foreach my $action ( sort keys %{ $commands{$object} } ) {
                my $action_key1
                    = exists $CLI_ORDER{$object}{$action}
                    ? $action
                    : q{default};
                my @cli_order = @{ $CLI_ORDER{$object}{$action_key1} };
                my $cli = join ( q{ }, map { q{$} . $_ } @cli_order );
                print qq{\tefs $action $object $cli };
                if ( exists $CLI_OPTIONS{$object} ) {
                    my $action_key2
                        = exists $CLI_OPTIONS{$object}->{$action}
                        ? $action
                        : q{default};
                    if ( ref $CLI_OPTIONS{$object}->{$action_key2}
                        && scalar @{ $CLI_OPTIONS{$object}->{$action_key2} }
                        > 0 ) {
                        print q{[ },
                            join ( q{ },
                            map { q{-} . $_ }
                                @{ $CLI_OPTIONS{$object}->{$action_key2} } ),
                            q{ ]};
                    }
                }
                print qq{\n};
            }
            print qq{\n};
        }
        return;
    }

    if ( $mode eq q{efs help object} ) {

        # This takes care of 'efs help object'
        print qq{\n};
        if ( defined $object_type ) {
            foreach my $action ( sort keys %{ $commands{$object_type} } ) {
                my $action_key1
                    = exists $CLI_ORDER{$object_type}{$action}
                    ? $action
                    : q{default};
                my @cli_order = @{ $CLI_ORDER{$object_type}{$action_key1} };
                my $cli = join ( q{ }, map { q{$} . $_ } @cli_order );
                print qq{\tefs $action $object_type $cli };
                if ( exists $CLI_OPTIONS{$object_type} ) {
                    my $action_key2
                        = exists $CLI_OPTIONS{$object_type}->{$action}
                        ? $action
                        : q{default};
                    if (scalar @{ $CLI_OPTIONS{$object_type}->{$action_key2} }
                        > 0 ) {
                        print q{[ },
                            join (
                            q{ },
                            map { qq{-$_} } @{
                                $CLI_OPTIONS{$object_type}->{$action_key2}
                            }
                        ),
                            q{ ]};
                    }
                }
                print qq{\n};
            }

        }
        print qq{\n};
        return;
    }

    if ( $mode eq q{efs action object --help} ) {

        # This takes care of 'efs action object -help'
        print qq{\n};
        foreach my $object ( sort keys %commands ) {
            next if ( $object ne $object_type );
            foreach my $a ( sort keys %{ $commands{$object} } ) {
                next if ( $action ne $a );
                my $tmp_cli_order_action
                    = exists $CLI_ORDER{$object}->{$action}
                    ? $action
                    : q{default};
                my @cli_order
                    = @{ $CLI_ORDER{$object}{$tmp_cli_order_action} };
                print qq{\tefs $action $object };

                # This prints out what the user supplied
                foreach ( @{$argv} ) {
                    print $_ , $SPACE;
                    shift @cli_order;
                }

                # Rest of the cli
                my $cli = join ( $SPACE, map { q{$} . $_ } @cli_order );
                print qq{$cli };

                if ( exists $CLI_OPTIONS{$object} ) {
                    my $tmp_cli_options_action
                        = exists $CLI_OPTIONS{$object}->{$action}
                        ? $action
                        : q{default};
                    if (scalar
                        @{ $CLI_OPTIONS{$object}->{$tmp_cli_options_action} }
                        > 0 ) {
                        print q{[ },
                            join (
                            q{ },
                            map { qq{-$_} } @{
                                $CLI_OPTIONS{$object}
                                    ->{$tmp_cli_options_action}
                                }
                            ),
                            q{ ]};
                    }
                }
                print qq{\n};
            }
            print qq{\n};
        }
        return;
    }

}

sub sigint_handler {
    ##  no critic (MagicNumbers)
    delete $SIG{INT};
    STDERR->print(
        color_warn(
            ( ( q{=} x 30 ), q{ WARNING }, ( q{=} x 30 ), qq{\n} ),
            qq{Stopping client will NOT abort the server-side action.\n},
            qq{If you are sure you want to quit JUST the client, while\n},
            qq{leaving the server-side action in progress, press ^C again.\n},
            ( ( q{=} x 69 ), qq{\n} )
        )
    );
}

sub get_envvars {

    my $vars = {};

    foreach my $key ( keys %ENV ) {
        next unless $key =~ m/(^ EFS) | ^ ANSI_COLORS_DISABLED $/msx;
        $vars->{$key} = $ENV{$key};
    }

    return $vars;

}

sub get_servernames {

    my (%args) = @_;

    if ( -f q{/var/run/efsd.pid} ) {
        return fqdn;
    }

    my %servers = EFS::Config->parse_efsservers;

    my $servertype = $args{servertype} || EFS::Config->parse_servertype;

    if ( not $servertype ) {
        die(
            qq{Unable to determine servertype automatically.\n},
            qq{Specify -serverhost or -servertype explicitly.\n},
        );
    }

    my $cell;


    # Use abbrev to normalize shorthand like a.wfd
    my %cells = abbrev keys %servers;
    $cell = $cells{ $args{servercell} };

    if ( not $cell ) {
        die qq{No EFS cells found that match $args{servercell}\n};
    }

    my @servers = ();

    if ( not ref $servers{$cell} ) {
        die qq{No EFS servers defined for cell '$cell'\n};
    }

    if ( not ref $servers{$cell}->{$servertype} ) {
        die qq{No EFS servers defined for servertype '$servertype'\n};
    }

    foreach my $server ( keys %{ $servers{$cell}->{$servertype} } ) {
        push ( @servers, $server );
    }

    if ( not @servers ) {
        die(
            q{Unable to determine server list for cell },
            qq{$cell, servertype $servertype\n}
        );
    }

    return shuffle(@servers);

}

sub get_serverport {

    Readonly my $PORT_DEFAULT => 17_831;

    if ( my @efsd = getservbyname ( q{efsd}, q{tcp} ) ) {
        return $efsd[2];
    }
    elsif ( -f q{/etc/efs/client.conf} ) {
        my $config = EFS::Config::Client->new;
        return ( $config->get_value( q{daemon}, q{port} ) || $PORT_DEFAULT );
    }
    else {
        return $PORT_DEFAULT;
    }

}

sub get_authuser_password {

    my $class = shift;
    my %args = @_;

    if ( $ENV{EFS_CLI_PASSWORD} ) {
        if ( $args{invalid_password} ) {
            die qq{Invalid authuser password (from EFS_CLI_PASSWORD)\n};
        }
        return $ENV{EFS_CLI_PASSWORD};
    }
    elsif ( -t STDIN ) {

        my $authpass = read_password( qq{Password for $args{authuser}: } );

        if ( not $authpass ) {
            die qq{Access denied\n};
        }

        return $authpass;

    }

}

{

    my $invalid_home_access = 0;

    sub verify_home_access {

        my $class = shift;
        my $authuser = shift;
        my $home = shift;

        if ( $invalid_home_access ) {
            return;
        }

        my $stat = stat($home);

        if ( not $stat ) {
            warn qq{Unable to stat $home: $ERRNO\n};
            $invalid_home_access++;
            return;
        }

        if ( $stat->uid != $EUID ) {
            warn qq{Directory $home is not owned by $authuser\n};
            $invalid_home_access++;
            return;
        }

        if ( $stat->mode & 022 ) {
            warn qq{Directory $home has invalid permissions\n};
            $invalid_home_access++;
            return;
        }

        return 1;

    }

}

sub get_cached_password {

    my $class = shift;
    my %args = @_;

    my $authuser = $args{authuser};
    my $home = $args{home};

    Readonly my $configdir => qq{$home/.efsconfig};
    Readonly my $authfile  => qq{$configdir/auth};

    if ( not -d $home ) {
        return;
    }

    if ( not $class->verify_home_access($authuser,$home) ) {
        return;
    }

    if ( not -d $configdir ) {
        return;
    }

    my $stat = stat($configdir);

    if ( not $stat ) {
        warn qq{Unable to stat $configdir: $ERRNO\n};
        return;
    }

    if ( $stat->uid != $EUID ) {
        warn qq{Directory $configdir is not owned by $authuser\n};
        return;
    }

    if ( $stat->mode & 077 ) {
        warn qq{Directory $configdir has invalid permissions\n};
        return;
    }

    if ( not -f $authfile ) {
        return;
    }

    my $config = Config::Tiny->read( $authfile );

    if ( not $config ) {
        warn qq{Unable to parse $authfile\n};
        return;
    }

    if ( $config->{security}->{password} ) {
        return EFS::CLI->rotate13( $config->{security}->{password} );
    }
    else {
        return;
    }

}

sub set_cached_password {

    my $class = shift;
    my %args = @_;

    my $authuser = $args{authuser};
    my $authpass = $args{authpass};
    my $home = $args{home};

    if ( not -d $home or not -w $home ) {
        return;
    }

    if ( not $class->verify_home_access($authuser,$home) ) {
        return;
    }

    Readonly my $configdir => qq{$home/.efsconfig};
    Readonly my $authfile  => qq{$configdir/auth};

    my $eval_error;
    try {
        EFS::Util->MakeDir( path => $configdir, mode => 0700 );
    }
    catch {
        $eval_error = $_;
    };

    if ( $eval_error ) {
        warn($eval_error);
        return;
    }

    my $authio = IO::File->new( $authfile, O_RDWR | O_CREAT, 0600 );

    if ( not $authio ) {
        warn qq{Unable to open $authfile: $ERRNO\n};
        return;
    }

    my $config = Config::Tiny->read( $authfile );

    if ( not $config ) {
        warn qq{Unable to parse $authfile: $ERRNO\n};
        return;
    }

    if ( $config->{security}->{q{no-auth-cache}} or $args{q{no-auth-cache}} ) {
        return;
    }

    warn qq{WARNING: Password cached in $authfile with light encryption\n};

    $config->{security}->{password} = EFS::CLI->rotate13( $authpass );

    $config->write( $authfile ) ||
        warn qq{Unable to update $home/.efsconfig/auth: $ERRNO\n};

    return 1;

}

sub get_username_password {

    my $class = shift;
    my %args = @_;

    my @getpwuid = getpwuid ($EUID)
        or die qq{Unable to getpwuid $EUID: ERRNO\n};

    my $authuser = $getpwuid[0];
    my $home = $ENV{HOME} || $getpwuid[7];

    # XXX: Temporary test environment hack (wpm, 1/25/10)
    if ( $authuser eq q{efsops} and -f q{/etc/efs/test.conf} ) {
        return qw(efsops efsops);
    }

    my ($authpass) = $class->get_cached_password(
        authuser => $authuser,
        home     => $home,
    );

    if ( $args{invalid_password} or not $authpass ) {
        $authpass = $class->query_password(
            prompt           => qq{Password for $authuser: },
            authuser         => $authuser,
            home             => $home,
            q{no-auth-cache} => $args{q{no-auth-cache}},
        );
    }

    if ( not $authpass ) {
        die qq{Access denied\n};
    }

    return ($authuser, $authpass);

}

sub query_password {

    my $class = shift;
    my %args = @_;

    if ( $ENV{EFS_TEST_FAKE_PASSWORD} ) {
        return $ENV{EFS_TEST_FAKE_PASSWORD};
    }

    my $authpass = read_password($args{prompt});

    if ( $authpass and not $args{q{no-auth-cache}} ) {
        $class->set_cached_password(
            authuser         => $args{authuser},
            authpass         => $authpass,
            home             => $args{home},
            q{no-auth-cache} => $args{q{no-auth-cache}},
        );
    }

    return $authpass;

}

sub rotate13 {
    my $class = shift;
    my $string = shift;
    $string =~ tr /a-zA-Z/n-za-mN-ZA-M/;
    return $string;
}

sub assert_platform_unused {

    my $class = shift;
    my $name = shift;

    my $type = $class->class2type;

    my %rs_of = (
        platforminstancealias =>
            $class->get_resultset( q{PlatformInstanceAlias} ),
        platformalias    => $class->get_resultset( q{PlatformAlias} ),
        platform         => $class->get_resultset( q{Platform} ),
        platforminstance => $class->get_resultset( q{PlatformInstance} ),
    );

    foreach my $othertype ( keys %rs_of ) {

        if ( $type eq $othertype ) {
            next;
        }

        if ( $rs_of{ $othertype }->search({ name => $name })->count ) {
            die qq{Invalid argument: $name is already a $othertype\n};
        }

    }

    return 1;

}

1;

# Local Variables:
#   mode: cperl
#   cperl-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab shiftwidth=4:
