package EFS::CLI::ChildRelease;

use strict;
use warnings;
use English;

use base q{EFS::CLI};
use base q{EFS::DBI};

use Params::Validate qw( :all );

sub create {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        metaproj => { type       => SCALAR },
        project  => { type       => SCALAR },
        name     => { type       => SCALAR },
        parent   => { type       => SCALAR },
    });

    my $project_rs = $class->get_resultset( q{Project} );
    my $project = $project_rs->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        name     => $args{project},
    );

    return EFS::RPC->execute(
        cell   => $project->cell,
        class  => $class,
        method => q{create_rpc},
        args   => [ %args ],
    );

}

sub create_rpc {

    my $class = shift;
    my %args  = @_;

    my $release_rs = $class->get_resultset( q{Release} );
    my $parent = $release_rs->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        project  => $args{project},
        name     => $args{parent},
    );

    if ( not $parent->locked ) {
        die qq{Parent $parent is not locked\n};
    }

    if ( not $parent->has_prod_stages ) {
        die qq{Parent $parent is not prod\n};
    }

    my $child = $release_rs->retrieve_bynames(
        metaproj => $args{metaproj},
        project  => $args{project},
        name     => $args{name},
    );

    if ( $child ) {

        if ( $child->locked ) {
            die qq{Child release $child already exists, and is locked\n};
        }

        my $history_rs = $class->get_resultset( q{History} );
        my $history = $history_rs->search({
            command  => q{create childrelease},
            metaproj => $args{metaproj},
            project  => $args{project},
            release  => $args{name},
        });

        if ( not $history->count ) {
            die qq{Release $child was not create as a child release\n};
        }

    }
    else {

        $child = EFS::CLI::Release->create(
            history  => $args{history},
            metaproj => $args{metaproj},
            project  => $args{project},
            name     => $args{name},
        );

    }

    my %parent_installs = map { $_->name => $_ } $parent->installs;

    foreach my $name ( sort keys %parent_installs ) {

        my $parent_install = $parent_installs{ $name };

        my $child_install = EFS::CLI::Install->create(
            history  => $args{history},
            metaproj => $args{metaproj},
            project  => $args{project},
            release  => $args{name},
            name     => $name,
        );

        EFS::Util->CreateLinkedDirectory(
            from => $child_install->get_path(q{dev}),
            to   => $parent_install->get_path(q{dist}),
        );

        EFS::CLI::Install->_lock_unlock(
            history  => $args{history},
            metaproj => $args{metaproj},
            project  => $args{project},
            release  => $args{name},
            name     => $name,
            lockmode => q{lock},
        );

    }

    EFS::CLI::Dependency->define(
        history            => $args{history},
        dependent_metaproj => $args{metaproj},
        dependent_project  => $args{project},
        dependent_release  => $args{name},
        metaproj           => $args{metaproj},
        project            => $args{project},
        release            => $args{parent},
        type               => q{runtime},
    );

    STDOUT->print( qq{Created child $child, of parent $parent\n} );

    return 1;

}

1;

# Local Variables:
#   mode: cperl
#   cperl-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab shiftwidth=4:
