package EFS::CLI::Dependency;

use strict;
use warnings;
use English;

use base q{EFS::CLI};
use base q{EFS::DBI};
use Params::Validate qw( :all );
use Try::Tiny;

sub history {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        %EFS::CLI::History::validation_specs,
    });

    return EFS::CLI::History->history(%args);

}

sub audit {

    my $class = shift;

    my @destroyed = ();

    my $resultset = $class->get_resultset( q{DependencyRelease} );

    foreach my $dependency_release ( $resultset->all ) {
        if ( $dependency_release->release->destroyed ){
            push(@destroyed,$dependency_release);
        }
    }

    if ( @destroyed ) {
        STDOUT->print(
            q{The following dependencies are invalid, },
            qq{because the release has been destroyed:\n}
        );
        $class->display_objects( objects => \@destroyed );
        return;
    } else {
        STDOUT->print( qq{No dependencies on destroyed releases\n} );
        return 1;
    }

}

sub auditcircular {

    my $class = shift;

    my $releasealias_rs = $class->get_resultset( q{ReleaseAlias} );
    my @releasealiases  = $releasealias_rs->all;

    my $errors = 0;

    my $dependency_release_rs      = $class->get_resultset( q{DependencyRelease} );
    my $dependency_releasealias_rs = $class->get_resultset( q{DependencyReleaseAlias} );

    foreach my $type ( qw( runtime build test ) ) {

        STDOUT->print( qq{Retrieving $type dependencies...\n} );

        my @dependencies = (
            $dependency_release_rs->search({ type => $type }),
            $dependency_releasealias_rs->search({ type => $type }),
        );

        STDOUT->print( qq{Building relationship hashes...\n} );

        my %dependencies = ();
        my %dependents   = ();

        foreach my $dependency ( @dependencies ) {

            my $dependent_mpr  = $dependency->dependent_object_id;
            my $dependency_mpr = $dependency->dependency_object_id;

            $dependencies{ $dependent_mpr }->{ $dependency_mpr }++;
            $dependents{ $dependency_mpr }->{ $dependent_mpr }++;

        }

        foreach my $releasealias ( @releasealiases ) {

            my $dependent_mpr  = $releasealias->object_id;
            my $dependency_mpr = $releasealias->target->object_id;

            $dependencies{ $dependent_mpr }->{ $dependency_mpr }++;
            $dependents{ $dependency_mpr }->{ $dependent_mpr }++;

        }

        try {
            $class->check_circular(
                dependencies => \%dependencies,
                dependents   => \%dependents,
            );
        } catch {
            my @errors = @_;
            warn @errors;
            $errors++;
        }

    }

    if ( $errors ) {
        die qq{Circular dependencies exist!!\n};
    } else {
        STDOUT->print( qq{No circular dependencies found\n} );
    }

    return 1;

}

sub define {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        dependent_metaproj => { type => SCALAR },
        dependent_project  => { type => SCALAR },
        dependent_release  => { type => SCALAR },
        metaproj           => { type => SCALAR },
        project            => { type => SCALAR },
        release            => { type => SCALAR },
        type               => { type => SCALAR },
        nocheckstage       => { type => SCALAR, optional => 1 },
    });

    my $release_rs = $class->get_resultset( q{Release} );

    my $dependent_release = $release_rs->retrieve_bynames_assert(
        metaproj => $args{dependent_metaproj},
        project  => $args{dependent_project},
        name     => $args{dependent_release},
    );

    my $dependent_mpr = $dependent_release->object_id;

    my $dependent_stage = $dependent_release->stage;

    if ( not $args{nocheckstage} and
             $dependent_release->has_only_destroyable_stages ) {
        die(
            qq{Release '$dependent_mpr' has stage '$dependent_stage'.\n},
            qq{Dependencies may not be defined for $dependent_stage releases.\n}
        );
    }

    my $dependency_object = $release_rs->retrieve_release_or_alias_assert(
        metaproj => $args{metaproj},
        project  => $args{project},
        name     => $args{release},
    );

    my $dependency_mpr = $dependency_object->object_id;

    my $dependency_tree = EFS::DependencyTree->new({
        seeds => [ $dependency_object ],
        type  => $args{type},
    });

    my $dependency_release_rs      = $class->get_resultset( q{DependencyRelease} );
    my $dependency_releasealias_rs = $class->get_resultset( q{DependencyReleaseAlias} );

    my $dependency_rs;
    my $dependency_key;

    if ( $dependency_object->isa(q{EFS::DBI::Result::Release}) ) {
        $dependency_rs  = $dependency_release_rs;
        $dependency_key = q{release};
    } else {
        $dependency_rs  = $dependency_releasealias_rs;
        $dependency_key = q{releasealias};
    }

    if ( $dependency_mpr eq $dependent_mpr ) {
        die(
            q{Invalid dependency specification.  },
            qq{Self-referential dependencies are not allowed.\n}
        );
    }

    if ( $dependency_tree->has_dependency( $dependent_mpr ) ) {
        STDOUT->print(
            qq{The dependent ($dependent_mpr) is },
            qq{one of the dependencies of $dependency_mpr:\n},
            $dependency_tree->get_tree_text(
                highlight => { $dependent_mpr => q{circular} },
            ),
            qq{This would create a circular dependency.\n},
        );
        die qq{Invalid dependency specification.\n};
    }

    if ( $args{type} eq q{runtime} and
             $dependent_stage eq q{prod} and
                 not $dependency_object->has_prod_stages ) {
        my $dependency_stage = $dependency_object->stage;
        die(
            qq{Invalid dependency specification.\n},
            qq{Dependent '$dependent_mpr' is already in stage 'prod', but\n},
            qq{dependency '$dependency_mpr' is in stage '$dependency_stage'\n},
            q{Once a release is 'prod', you can not define },
            qq{non-prod runtime dependencies.\n},
        );
    }

    my %create_args = (
        dependent_release => $dependent_release->id,
        $dependency_key   => $dependency_object->id,
        type              => $args{type},
    );

    my ($dependency) = $dependency_rs->search( \%create_args );

    if ( $dependency ) {
        STDOUT->print(
            qq{Verified dependency of $dependent_mpr on $dependency_mpr, },
            qq{type $args{type}\n}
        );
    } else {
        $dependency = $dependency_rs->create( \%create_args );
        STDOUT->print(
            qq{Defined dependency of $dependent_mpr on $dependency_mpr, },
            qq{type $args{type}\n}
        );
    }

    return $dependency;

}

sub undef {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        dependent_metaproj => { type => SCALAR },
        dependent_project  => { type => SCALAR },
        dependent_release  => { type => SCALAR },
        metaproj           => { type => SCALAR },
        project            => { type => SCALAR },
        release            => { type => SCALAR },
        type               => { type => SCALAR },
    });

    my $release_rs = $class->get_resultset( q{Release} );

    my $dependent_release = $release_rs->retrieve_bynames_assert(
        metaproj => $args{dependent_metaproj},
        project  => $args{dependent_project},
        name     => $args{dependent_release},
    );

    my $dependent_mpr = $dependent_release->object_id;

    my $dependency_object = $release_rs->retrieve_release_or_alias_assert(
        metaproj => $args{metaproj},
        project  => $args{project},
        name     => $args{release},
    );

    my $dependency_mpr = $dependency_object->object_id;

    my $dependency_rs;
    my $dependency_key;

    if ( $dependency_object->isa(q{EFS::DBI::Result::ReleaseAlias}) ) {
        $dependency_rs  = $class->get_resultset( q{DependencyReleaseAlias} );
        $dependency_key = q{releasealias};
    } else {
        $dependency_rs  = $class->get_resultset( q{DependencyRelease} );
        $dependency_key = q{release};
    }

    my ($dependency) = $dependency_rs->search({
        dependent_release => $dependent_release->id,
        $dependency_key   => $dependency_object->id,
        type              => $args{type},
    });

    if ( not $dependency ) {
        die(
            qq{No such $args{type} dependency of $dependent_mpr },
            qq{on $dependency_mpr\n}
        );
    }

    $dependency->delete;

    STDOUT->print(
        qq{Undefined dependency of $dependent_mpr on $dependency_mpr\n}
    );

    return 1;

}

#
# efs batchdefine dependency foo bar 1.0 runtime a/b/c,d/e/f,g/h/i
#
sub batchdefine {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        type               => { type => SCALAR },
        dependent_metaproj => { type => SCALAR },
        dependent_project  => { type => SCALAR },
        dependent_release  => { type => SCALAR },
        dependencies       => { type => SCALAR},
        complete           => { type => BOOLEAN, optional => 1 },
    });

    my $release_rs = $class->get_resultset( q{Release} );

    my $dependent_release = $release_rs->retrieve_bynames_assert(
        metaproj => $args{dependent_metaproj},
        project  => $args{dependent_project},
        name     => $args{dependent_release},
    );

    my $dependent_mpr   = $dependent_release->object_id;
    my $dependent_stage = $dependent_release->stage;

    if ( $dependent_release->has_only_destroyable_stages ) {
        die(
            qq{Release '$dependent_mpr' has stage '$dependent_stage'.\n},
            qq{Dependencies may not be defined for $dependent_stage releases.\n}
        );
    }

    # Strip any dangling leading/trailing separators
    $args{dependencies} =~ s/\A [\s,]+//msx;
    $args{dependencies} =~ s/[\s,]+ \z//msx;

    my @dependencies = split( m/[\s,]+/msx, $args{dependencies} );

    my $selfref = q{};

    my %invalid  = ();
    my %nonprod  = ();
    my %circular = ();

    my %releases       = ();
    my %releasealiases = ();

  DEPENDENCY: foreach my $dependency ( @dependencies ) {

        if ( $dependent_mpr eq $dependency ) {
            $selfref++;
            next DEPENDENCY;
        }

        my ($metaproj_name,$project_name,$release_name) =
            split( m{ / }msx, $dependency );

        my $dependency_object = $release_rs->retrieve_release_or_alias(
            metaproj => $metaproj_name,
            project  => $project_name,
            name     => $release_name,
        );

        if ( not $dependency_object ) {
            $invalid{$dependency}++;
            next DEPENDENCY;
        }

        if ( $dependency_object->isa(q{EFS::DBI::Result::Release}) ) {
            $releases{$dependency} = $dependency_object;
        } else {
            $releasealiases{$dependency} = $dependency_object;
        }

        my $dependency_tree = EFS::DependencyTree->new({
            seeds => [ $dependency_object ],
            type  => $args{type},
        });

        if ( $dependency_tree->has_dependency( $dependent_mpr ) ) {
            $circular{$dependency} = $dependency_tree;
        }

        if ( $args{type} eq q{runtime} && $dependent_stage eq q{prod} ) {
            if ( not $dependency_object->has_prod_stages ) {
                $nonprod{$dependency} = $dependency_object->stage;
            }
        }

    }

    my @fatal = ();

    if ( $selfref ) {
        push @fatal, qq{Self-referential dependencies are not allowed.\n};
    }

    if ( %invalid ) {
        push @fatal, (
            qq{Must be releases or releasealiases:\n\t},
            join(qq{\n\t}, sort { lc($a) cmp lc($b) } keys %invalid), qq{\n}
        );
    }

    if ( %nonprod ) {
        push @fatal, (
            qq{Dependent '$dependent_mpr' is already in stage 'prod'.\n},
            q{The following runtime dependencies are },
            qq{not prod or deprecated:\n\t},
            join(qq{\n\t}, map { qq{$_ => $nonprod{$_}} } keys %nonprod ), qq{\n},
        );
    }

    if ( %circular ) {
        push @fatal, (
            qq{Circular dependencies are not allowed.\n},
            qq{Dependent '$dependent_mpr' is a dependency of:\n},
        );
        foreach my $dependency ( sort { lc($a) cmp lc($b) } keys %circular ) {
            push @fatal, (
                $circular{$dependency}->get_tree_text(
                    highlight => { $dependent_mpr => q{circular} },
                ),
            );
        }
    }

    if ( @fatal ) {
        die( qq{Invalid dependencies specified.\n}, @fatal );
    }

    foreach my $dependency_mpr ( sort { lc($a) cmp lc($b) } @dependencies ) {

        my ($metaproj_name,$project_name,$release_name) =
            split( m{ / }msx, $dependency_mpr );

        my $dependency_key;
        my $dependency_rs;
        my $dependency_object;

        if ( $releases{$dependency_mpr} ) {
            $dependency_key    = q{release};
            $dependency_object = $releases{$dependency_mpr};
            $dependency_rs     =
                $class->get_resultset( q{DependencyRelease} );
        } else {
            $dependency_key    = q{releasealias};
            $dependency_object = $releasealiases{$dependency_mpr};
            $dependency_rs     =
                $class->get_resultset( q{DependencyReleaseAlias} );
        }

        my %create_args = (
            dependent_release => $dependent_release->id,
            $dependency_key   => $dependency_object->id,
            type              => $args{type},
        );

        my ($dependency) = $dependency_rs->search( \%create_args );

        if ( $dependency ) {
            STDOUT->print(
                qq{Verified $args{type} dependency of },
                qq{$dependent_mpr on $dependency_mpr\n}
            );
        } else {
            $dependency_rs->create( \%create_args );
            STDOUT->print(
                qq{Defined $args{type} dependency of },
                qq{$dependent_mpr on $dependency_mpr\n}
            );
        }

    }

    if ( not $args{complete} ) {
        STDOUT->print( qq{All dependencies verified or defined as specified.\n} );
        return 1;
    }

    my %search_args = (
        dependent_release => $dependent_release->id,
        type              => $args{type},
    );

    foreach my $subclass ( qw( DependencyRelease DependencyReleaseAlias ) ) {

        my $dependency_rs = $class->get_resultset( $subclass );

        my @dependencies =
            $class->sort_objects( $dependency_rs->search( \%search_args ) );

        foreach my $dependency ( @dependencies ) {

            my $dependency_mpr = $dependency->dependency_object_id;

            next if $releases{$dependency_mpr};

            next if $releasealiases{$dependency_mpr};

            $dependency->delete;

            STDOUT->print(
                qq{Undefined unspecified $args{type} dependency of },
                qq{$dependent_mpr on $dependency_mpr\n}
            );

        }

    }

    STDOUT->print(
        qq{Complete dependency specification verified and/or defined\n}
    );

    return 1;

}

sub batchundef {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        type               => { type => SCALAR },
        dependent_metaproj => { type => SCALAR },
        dependent_project  => { type => SCALAR },
        dependent_release  => { type => SCALAR },
        dependencies       => { type => SCALAR },
    });

    my $release_rs = $class->get_resultset( q{Release} );

    my $dependent_release = $release_rs->retrieve_bynames_assert(
        metaproj => $args{dependent_metaproj},
        project  => $args{dependent_project},
        name     => $args{dependent_release},
    );

    my $dependent_mpr = $dependent_release->object_id;

    my @dependencies = split( m/[\s,]+/msx, $args{dependencies} );

    foreach my $dependency_mpr ( @dependencies ) {

        my ($metaproj_name,$project_name,$release_name) =
            split( m{ / }msx, $dependency_mpr );

        my $dependency_object = $release_rs->retrieve_release_or_alias_assert(
            metaproj => $metaproj_name,
            project  => $project_name,
            name     => $release_name,
        );

        my $dependency_key;
        my $dependency_rs;

        if ( $dependency_object->isa(q{EFS::DBI::Result::Release}) ) {
            $dependency_key = q{release};
            $dependency_rs  =
                $class->get_resultset( q{DependencyRelease} );
        } else {
            $dependency_key = q{releasealias};
            $dependency_rs  =
                $class->get_resultset( q{DependencyReleaseAlias} );
        }

        my ($dependency) = $dependency_rs->search({
            dependent_release => $dependent_release->id,
            $dependency_key   => $dependency_object->id,
            type              => $args{type},
        });

        if ( $dependency ) {
            $dependency->delete;
            STDOUT->print(
                qq{Undefined $args{type} dependency of $dependent_mpr },
                qq{on $dependency_mpr\n}
            );
        } else {
            STDOUT->print(
                qq{No such $args{type} dependency of },
                qq{$dependent_mpr on $dependency_mpr\n}
            );
        }

    }

    STDOUT->print(
        qq{All specified dependencies verified as missing or undefined.\n}
    );

    return 1;

}

sub diststate {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        metaproj => { type => SCALAR },
        project  => { type => SCALAR },
        release  => { type => SCALAR },
        %EFS::CLI::validate_diststate_args,
    });

    my $release_rs = $class->get_resultset( q{Release} );

    my $release = $release_rs->retrieve_bynames_assert(
        metaproj => $args{metaproj},
        project  => $args{project},
        name     => $args{release},
    );
    my $mpr_name = $release->object_id;

    my @cells = $class->handle_cell_arguments(
        %args,
        include_parents => 1,
        default_scope   => q{global},
    );

    my @prod_cells = grep { $_->celltype eq q{prod} } @cells;

    my @dependencies = $release->dependencies( type => q{runtime} );

    if ( not @dependencies ) {
        STDOUT->print(
            qq{The specified release $mpr_name has no runtime },
            qq{dependencies defined.\n},
        );
        return 1;
    }

    my $errors = 0;

    foreach my $dependency ( @dependencies ) {

        my $diststate        = $dependency->get_diststate;
        my $dependency_mpr   = $dependency->dependency_object_id;
        my $dependency_stage = $dependency->stage;
        my $dependency_type  =
            $dependency->isa(q{EFS::DBI::Result::DependencyRelease}) ?
                q{release} : q{releasealias};

        if ( @prod_cells and not
                 $dependency->target->has_prod_stages ) {
            warn(
                qq{Dependency $dependency_mpr has stage $dependency_stage, },
                qq{can't be disted to prod cells\n}
            );
            $errors++;
        } elsif ( $diststate->cells_needing_update( cells => \@cells ) ) {
            warn(
                qq{Diststate for $dependency_type dependency $dependency_mpr:\n}
            );
            $diststate->display_text_table( cells => \@cells );
            $errors++;
        } else {
            STDOUT->print(
                qq{Dependency $dependency_mpr is up to date in },
                qq{all requested cells\n},
            );
        }

    }

    if ( $errors ) {
        die qq{One or more dependencies are not up to date.\n};
    } else {
        return 1;
    }

    return;
}

sub showtree {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        metaproj  => { type => SCALAR },
        project   => { type => SCALAR },
        release   => { type => SCALAR },
        type      => { type => SCALAR, default  => q{runtime}, optional => 1 },
        kind      => { type => SCALAR, default  => q{dependencies}, optional => 1 },
        maxdepth  => { type => SCALAR, default  => 0, optional => 1 },
        conflicts => { type => BOOLEAN, optional => 1 },
        %EFS::CLI::validate_dist_args,
    });

    my @cells = $class->handle_cell_arguments(
        %args,
        include_parents => 1,
        default_scope   => q{global},
    );

    my $release_rs = $class->get_resultset( q{Release} );

    my $object = $release_rs->retrieve_release_or_alias_assert(
        metaproj => $args{metaproj},
        project  => $args{project},
        name     => $args{release},
    );

    my $tree = EFS::DependencyTree->new({
        seeds => [ $object ],
        type  => $args{type},
        kind  => $args{kind},
    });

    my $project_rs  = $class->get_resultset( q{Project} );
    my $metaproj_rs = $class->get_resultset( q{Metaproj} );

    if ( $args{conflicts} ) {

        my %conflicts = ();

        if ( $args{conflicts} eq q{all} ) {

            %conflicts = $tree->find_conflicts;

        } else {

            my %interested = ();

            foreach my $interested ( split( m/[\s,]+/msx, $args{conflicts} ) ) {

                my @fields = split( m{ / }msx, $interested );

                if ( scalar @fields > 2 ) {
                    die(
                        qq{Invalid conflicts specification.\n},
                        qq{Only metaprojs or projects may be specified.\n},
                        qq{The value '$interested' is invalid.\n},
                    );
                }

                my ($metaproj_name,$project_name) = @fields;

                if ( $project_name ) {
                    $project_rs->retrieve_bynames_assert(
                        metaproj => $metaproj_name,
                        name     => $project_name,
                    );
                } else {
                    $metaproj_rs->retrieve_bynames_assert(
                        name => $metaproj_name,
                    );
                }

                $interested{ $interested }++;

            }

            %conflicts = $tree->find_conflicts( interested => \%interested );

        }

        STDOUT->print(
            $tree->get_tree_text(
                maxdepth  => $args{maxdepth},
                highlight => \%conflicts,
                cells     => \@cells,
            ),
        );

    } else {
        STDOUT->print(
            $tree->get_tree_text(
                maxdepth => $args{maxdepth},
                cells    => \@cells,
            ),
        );
    }

    return 1;

}

sub display {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        dependent_metaproj => { type     => SCALAR,
                                optional => 1 },
        dependent_project  => { type     => SCALAR,
                                depends  => [qw( dependent_metaproj )],
                                optional => 1 },
        dependent_release  => { type     => SCALAR,
                                depends  => [qw( dependent_metaproj
                                                 dependent_project )],
                                optional => 1 },
        metaproj           => { type     => SCALAR,
                                optional => 1 },
        project            => { type     => SCALAR,
                                depends  => [qw( metaproj )],
                                optional => 1 },
        release            => { type     => SCALAR,
                                depends  => [qw( metaproj project )],
                                optional => 1 },
        type               => { type     => SCALAR,
                                default  => q{runtime,build,test},
                                optional => 1 },
        direct             => { type     => SCALAR,
                                optional => 1 },
    });

    my @types = split m{[,\s]+}msx, $args{type};

    my $metaproj_rs     = $class->get_resultset( q{Metaproj} );
    my $project_rs      = $class->get_resultset( q{Project} );
    my $release_rs      = $class->get_resultset( q{Release} );
    my $releasealias_rs = $class->get_resultset( q{ReleaseAlias} );

    my $dependency_release_rs      = $class->get_resultset( q{DependencyRelease} );
    my $dependency_releasealias_rs = $class->get_resultset( q{DependencyReleaseAlias} );

    my %search = ();

    if ( $args{dependent_release} ) {

        my $releasealias = $releasealias_rs->retrieve_bynames(
            metaproj => $args{dependent_metaproj},
            project  => $args{dependent_project},
            name     => $args{dependent_release},
        );

        if ( $releasealias ) {
            my $target_release = $releasealias->target_release;
            STDOUT->print(
                qq{Normalizing releasealias ($releasealias) to its },
                qq{target release ($target_release)\n}
            );
            $search{dependent_release} = [ $target_release->id ];
        }
        else {
            my $release = $release_rs->retrieve_bynames_assert(
                metaproj => $args{dependent_metaproj},
                project  => $args{dependent_project},
                name     => $args{dependent_release},
            );
            $search{dependent_release} = [ $release->id ];
        }

    }
    elsif ( $args{dependent_project} ) {

        my $project = $project_rs->retrieve_bynames_assert(
            metaproj => $args{dependent_metaproj},
            name     => $args{dependent_project},
        );

        if ( my @ids = map { $_->id } $project->releases ) {
            $search{dependent_release} = \@ids,
        }

    }
    elsif ( $args{dependent_metaproj} ) {

        my $metaproj = $metaproj_rs->retrieve_bynames_assert(
            name => $args{dependent_metaproj},
        );

        if ( my @ids = map { $_->id } $metaproj->releases ) {
            $search{dependent_release} = \@ids,
        }

    }

    my $dependencies = {
        release      => [],
        releasealias => [],
    };

    if ( $args{release} ) {

        my ($object) = $release_rs->retrieve_release_or_alias_assert(
            metaproj => $args{metaproj},
            project  => $args{project},
            name     => $args{release},
        );

        my $object_type = $object->class2type;

        $dependencies->{$object_type} = [
            $object->dependency_resultset->search({
                %search,
                $object_type => $object->id,
                type         => \@types,
            })
        ];

        if ( not $args{direct} ) {
            if ( my @ids = map { $_->id } $object->all_releasealiases ) {
                push @{ $dependencies->{releasealias} },
                    $dependency_releasealias_rs->search({
                        %search,
                        releasealias => \@ids,
                        type         => \@types,
                    });
            }
        }

    } else {

        my %search_release      = %search;
        my %search_releasealias = %search;

        if ( $args{metaproj} ) {

            my $object;

            if ( $args{project} ) {
                $object = $project_rs->retrieve_bynames_assert(
                    metaproj => $args{metaproj},
                    name     => $args{project},
                );
            }
            else {
                $object = $metaproj_rs->retrieve_bynames_assert(
                    name => $args{metaproj},
                );
            }

            if ( my @ids = map { $_->id } $object->releases ) {
                $search_release{release} = \@ids;
            }
            if ( my @ids = map { $_->id } $object->releasealiases ) {
                $search_releasealias{releasealias} = \@ids;
            }

        }

        $dependencies->{release} = [
            $dependency_release_rs->search({
                %search_release,
                type => \@types,
            })
        ];

        $dependencies->{releasealias} = [
            $dependency_releasealias_rs->search({
                %search_releasealias,
                type => \@types,
            })
        ];

    }

    my @releases = @{ $dependencies->{release} };
    my @aliases  = @{ $dependencies->{releasealias} };

    if ( not @releases and not @aliases ) {
        STDOUT->print( qq{No dependencies match the specified criteria\n} );
        return 1;
    }

    if ( @releases ) {
        $releases[0]->display_objects(
            objects => [ $class->sort_objects( @releases ) ],
        );
    }

    if ( @releases and @aliases ) {
        STDOUT->print( qq{\n} );
    }

    if ( @aliases ) {
        $aliases[0]->display_objects(
            objects => [ $class->sort_objects( @aliases ) ],
        );

    }

    return 1;

}

1;

# Local Variables:
#   mode: cperl
#   cperl-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab shiftwidth=4:
