package EFS::DBI::Result::Fileqtree;

use strict;
use warnings;
use English;

use base q{EFS::DBI::Result};

__PACKAGE__->load_components( qw/+EFS::DBI::Prefix Core/ );

__PACKAGE__->table( q{fileqtree} );

__PACKAGE__->add_columns( qw( fileshare name ) );
__PACKAGE__->add_columns(
    id => { __PACKAGE__->get_id_column_attrs },
);
__PACKAGE__->set_primary_key( q{id} );
__PACKAGE__->sequence( q{fileqtree_seq} );

__PACKAGE__->belongs_to( fileshare => q{EFS::DBI::Result::Fileshare} );

__PACKAGE__->has_many( attrs => q{EFS::DBI::Result::FileqtreeAttrs} );

sub cell {
    return shift->fileshare->cell;
}

sub fileserver {
    return shift->fileshare->fileserver;
}

sub filetype {
    return shift->fileshare->filetype;
}

sub has_exectype {
    return ( shift->filetype ~~ [qw(dev dist)] );
}

sub assert_exectype {
    my $self = shift;
    my $filetype = $self->filetype;
    if ( not $self->has_exectype ) {
        die qq{Filetype is $filetype, not dev or dist\n};
    }
    return 1;
}

sub object_id {
    my $self = shift;
    my $separator = shift || q{/};
    my $object_id = join( $separator,
                          $self->fileshare->object_id,
                          $self->name );
    $object_id =~ s{/+}{/}gmsx;
    return $object_id;
}

sub admin_path {

    my $self = shift;

    my $pathname = join( q{/},
                         q{/.efsadmin},
                         $self->cell->name,
                         $self->fileserver->name,
                         $self->fileshare->name,
                         $self->name );

    $pathname =~ s{/+}{/}gmsx;

    return $pathname;

}

sub not_empty {

    my $self = shift;
    my $admin_path = $self->admin_path;
    my $directory = IO::Dir->new($admin_path) ||
        die qq{Unable to opendir $admin_path: $ERRNO\n};

    return grep { $_ ne q{.} && $_ ne q{..} } $directory->read;

}

sub is_empty {
    return not shift->not_empty;
}

sub get_execdirs {

    my $self = shift;

    my $admin_path = $self->admin_path;
    my $type       = $self->filetype;
    my $top        = qq{$admin_path/$type};

    my @execdirs = ();

    foreach my $metaproj ( EFS::Util->GetSubdirs( $top ) ) {
        foreach my $project ( EFS::Util->GetSubdirs( qq{$top/$metaproj} ) ) {
            foreach my $release ( EFS::Util->GetSubdirs( qq{$top/$metaproj/$project} ) ) {
                my $subdir = qq{$top/$metaproj/$project/$release};
                $subdir   .= q{/install} if $type eq q{dev};
                $subdir   .= q{/.exec};
                next if not -d $subdir;
                push @execdirs, $subdir;
            }
        }
    }

    return @execdirs;

}

1;

# Local Variables:
#   mode: cperl
#   cperl-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab shiftwidth=4:
