package EFS::DBI::ResultSet::Platform;

use strict;
use warnings;
use English;

use base q{EFS::DBI::ResultSet};

use Params::Validate qw( :all );

sub validate_data {

    my $class = shift;
    my %data  = @_;

    foreach my $type ( qw( platform platforminstance backwards emulates ) ) {
        next unless exists $data{$type};
        next if $data{$type} eq q{NULL};  # Special case
        ## no critic ProhibitEnumeratedClasses
        # we really do want a-z here; [:lower:] would be effected by locale
        $data{$type} =~ m/\A [-a-z0-9\.]+ \z/msx || die(
            qq{Invalid $type name '$data{$type}'\n},
            q{Must consist solely of lower case alphanumeric },
            qq{characters, dashes and/or dots\n}
        );
    }

    return;
}

sub retrieve_bynames_assert {

    my $self = shift;

    my (%args) = validate( @_, {
        name => { type => SCALAR },
    });

    my ($platform) = $self->search({ name => $args{name} });

    if ( not $platform ) {
        die qq{No such platform '$args{name}'\n};
    }

    return $platform;

}

sub get_flattened_data {

    my $self = shift;

    my $data = {};

    foreach my $platform ( $self->all ) {
        $data->{ $platform->name } = {
            backwards =>
                $platform->backwards ? $platform->backwards->name : q{},
            emulates  =>
                $platform->emulates  ? $platform->emulates->name  : q{},
        };
    }

    return $data;

}

sub get_roots {

    my $self = shift;

    my @roots = ();

    foreach my $platform ( $self->all ) {
        if ( not $platform->backwards and not $platform->emulates ) {
            push @roots, $platform;
        }
    }

    return @roots;

}

sub get_hierarchy_order {

    my $self = shift;

    my @roots = $self->get_roots;

    my $ordered = [ map { $_->name } @roots ];

    foreach my $root ( @roots ) {
        $self->get_children(
            name    => $root->name,
            ordered => $ordered,
        );
    }

    return $ordered;

}

sub get_children {

    my $self = shift;
    my %args = @_;

    my $name    = $args{name};
    my $ordered = $args{ordered};

    my ($platform) = $self->search({ name => $name });

    foreach my $alias ( $platform->aliases ) {
        push @{ $ordered }, $alias->name;
    }

    foreach my $instance ( $platform->instances ) {
        push @{ $ordered }, $instance->name;
        foreach my $alias ( $instance->aliases ) {
            push @{ $ordered }, $alias->name;
        }
    }

    foreach my $emulator ( $platform->emulators ) {
        push @{ $ordered }, $emulator->name;
        $self->get_children(
            name    => $emulator->name,
            ordered => $ordered,
        );
    }

    foreach my $forward ( $platform->forwards ) {
        push @{ $ordered }, $forward->name;
        $self->get_children(
            name    => $forward->name,
            ordered => $ordered,
        );
    }

    return 1;

}

1;

# Local Variables:
#   mode: cperl
#   cperl-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab shiftwidth=4:
