#!/usr/bin/env perl
use strict;
use warnings;
use English;

use feature q{switch};

use IO::File;
use Test::More;
use File::Basename qw( basename );

use blib;
use EFS;

use lib q{test/lib};
use EFS::Test::Config;

my $config = EFS::Test::Config->new;

my $fstype = $config->get_value( q{Filesystem}, q{type} );

my $sshcmd = $config->get_value( q{Global}, q{sshcmd} );

my $ldc_name       = $config->get_value( q{LocalDevCell}, q{name} );
my $ldc_efsserver  = $config->get_value( q{LocalDevCell}, q{efsserver} );
my $ldc_fileserver = $config->get_value( q{LocalDevCell}, q{fileserver} );
my $ldc_devshare   = $config->get_value( q{LocalDevCell}, q{devshare} );
my $ldc_homeshare  = $config->get_value( q{LocalDevCell}, q{homeshare} );
my $ldc_distshare  = $config->get_value( q{LocalDevCell}, q{distshare} );

my $rdc_name       = $config->get_value( q{RemoteDevCell}, q{name} );
my $rdc_efsserver  = $config->get_value( q{RemoteDevCell}, q{efsserver} );
my $rdc_fileserver = $config->get_value( q{RemoteDevCell}, q{fileserver} );
my $rdc_devshare   = $config->get_value( q{RemoteDevCell}, q{devshare} );
my $rdc_homeshare  = $config->get_value( q{RemoteDevCell}, q{homeshare} );
my $rdc_distshare  = $config->get_value( q{RemoteDevCell}, q{distshare} );

my $lpc_name       = $config->get_value( q{LocalProdCell}, q{name} );
my $lpc_efsserver  = $config->get_value( q{LocalProdCell}, q{efsserver} );
my $lpc_fileserver = $config->get_value( q{LocalProdCell}, q{fileserver} );
my $lpc_distshare  = $config->get_value( q{LocalProdCell}, q{distshare} );

if ( $fstype eq q{nfs} ) {
    if ( $ldc_efsserver ne $ldc_fileserver
             || $rdc_efsserver ne $rdc_fileserver
                   || $lpc_fileserver !~ /^fake/msx ) {
        plan skip_all => q{Unable to bootstrap the efsservers if they use NFS servers (yet)};
    }
}

my @localdirs   = qw( /efs );
my @remotedirs  = qw( /efs );
my @locallinks  = ();
my @remotelinks = ();

given ( $fstype ) {

    when ( q{afs} ) {

        @locallinks = (
            [ qq{/afs/$ldc_name/efs/global/dev},  q{/efs/dev} ],
            [ qq{/afs/$ldc_name/efs/global/home}, q{/efs/home} ],
            [ qq{/afs/$ldc_name/efs/dist},        q{/efs/dist} ],
        );

        @remotelinks = (
            [ qq{/afs/$rdc_name/efs/global/dev},  q{/efs/dev} ],
            [ qq{/afs/$rdc_name/efs/global/home}, q{/efs/home} ],
            [ qq{/afs/$rdc_name/efs/dist},        q{/efs/dist} ],
        );

    }

    when ( q{nfs} ) {

        @locallinks = (
            [ qq{$ldc_devshare/efs_qtree/dev},   q{/efs/dev} ],
            [ qq{$ldc_homeshare/efs_qtree/home}, q{/efs/home} ],
            [ qq{$ldc_distshare/efs_qtree/dist}, q{/efs/dist} ],
        );

        @remotelinks = (
            [ qq{$rdc_devshare/efs_qtree/dev},   q{/efs/dev} ],
            [ qq{$rdc_homeshare/efs_qtree/home}, q{/efs/home} ],
            [ qq{$rdc_distshare/efs_qtree/dist}, q{/efs/dist} ],
        );

        push @localdirs, (
            q{/.efs},
            q{/.efsglobal},
            qq{/.efsadmin/$ldc_name/$ldc_fileserver/var/tmp},
            qq{/.efsadmin/$lpc_name/$lpc_fileserver/var/tmp},
            qq{$ldc_devshare/efs_qtree/dev},
            qq{$ldc_homeshare/efs_qtree/home},
            qq{$ldc_distshare/efs_qtree/dist},
            qq{$lpc_distshare/efs_qtree/dist},
        );

        push @locallinks, (
            [ $ldc_devshare,  qq{/.efsadmin/$ldc_name/$ldc_fileserver/$ldc_devshare} ],
            [ $ldc_homeshare, qq{/.efsadmin/$ldc_name/$ldc_fileserver/$ldc_homeshare} ],
            [ $ldc_distshare, qq{/.efsadmin/$ldc_name/$ldc_fileserver/$ldc_distshare} ],
            [ $lpc_distshare, qq{/.efsadmin/$lpc_name/$lpc_fileserver/$lpc_distshare} ],
            [ $ldc_devshare,  q{/.efsglobal/} . basename($ldc_devshare) ],
            [ $ldc_homeshare, q{/.efsglobal/} . basename($ldc_homeshare) ],
        );

        push @remotedirs, (
            q{/.efs},
            q{/.efsglobal},
            q{/.efsadmin},
            qq{/.efsadmin/$rdc_name/$rdc_fileserver/var/tmp},
            qq{$rdc_devshare/efs_qtree/dev},
            qq{$rdc_homeshare/efs_qtree/home},
            qq{$rdc_distshare/efs_qtree/dist},
        );

        push @remotelinks, (
            [ $rdc_devshare,  qq{/.efsadmin/$rdc_name/$rdc_fileserver/$rdc_devshare} ],
            [ $rdc_homeshare, qq{/.efsadmin/$rdc_name/$rdc_fileserver/$rdc_homeshare} ],
            [ $rdc_distshare, qq{/.efsadmin/$rdc_name/$rdc_fileserver/$rdc_distshare} ],
            [ $rdc_devshare,  q{/.efsglobal/} . basename($rdc_devshare) ],
            [ $rdc_homeshare, q{/.efsglobal/} . basename($rdc_homeshare) ],
        );

    }

}

foreach my $localdir (@localdirs) {
    system( qq{mkdir -p $localdir} );
    $CHILD_ERROR and die qq{Unable to mkdir -p $localdir\n};
    ok( 1, qq{mkdir -p $localdir} );
}

foreach my $locallink (@locallinks) {
    my ($to,$from) = @{ $locallink };
    makelink( $to, $from );
    ok( 1, qq{ln -s $to $from} );
}

foreach my $remotedir (@remotedirs) {
    system( qq{$sshcmd $rdc_efsserver mkdir -p $remotedir} );
    $CHILD_ERROR and die qq{Unable to mkdir $rdc_efsserver:$remotedir\n};
    ok( 1, qq{$sshcmd $rdc_efsserver mkdir $remotedir} );
}

foreach my $remotelink (@remotelinks) {
    my ( $to, $from ) = @{ $remotelink };
    system (qq{$sshcmd $rdc_efsserver "bash -c '[[ -L $from ]] || ln -s $to $from'"});
    $CHILD_ERROR and die qq{Unable to create symlink $rdc_efsserver:$from -> $to\n};
    ok( 1, qq{$sshcmd $rdc_efsserver ln -s $to $from} );
}

done_testing();
exit 0;

sub makelink {

    my ( $to, $from ) = @_;

    if ( -l $from ) {
        my $readlink = readlink ($from) || die qq{Unable to readlink $from: $ERRNO\n};
        if ( $readlink eq $to ) {
            return 1;
        }
        unlink ($to) || die qq{Unable to unlink $to: $ERRNO\n};
    }

    symlink ( $to, $from ) || die qq{Unable to symlink $from -> $to: $ERRNO\n};

    return 1;

}

1;
