#!/usr/bin/env perl

use strict;
use warnings;
use English;

use Test::More tests => 54;
use Test::Exception;

use File::stat;

use blib;
use EFS;

my $testdir = qq{/var/tmp/EFS_Util_tests_$PID};

## no critic: RequireExtendedFormatting

# MakeDir

throws_ok {
    EFS::Util->MakeDir();
} qr{^Mandatory parameter.* missing in call to}ms,
    q{MakeDir parameter validation};

throws_ok {
    EFS::Util->MakeDir(
        path  => q{/this/will/surely/fail},
        user  => q{root},
        group => q{root},
        mode  => 0755,
    );
} qr{^Unable to mkdir /this/will/surely/fail:}ms,
    q{MakeDir failure};

ok(
    EFS::Util->MakeDir(
        path  => $testdir,
        user  => q{root},
        group => q{root},
        mode  => 0755,
    ),
    q{MakeDir success}
);

my $stat = stat($testdir) || die qq{stat: $ERRNO};

my ($root_uid) = 0;
my ($root_gid) = 0;

ok( $stat->uid() == $root_uid, q{MakeDir uid is correct} );
ok( $stat->gid() == $root_gid, q{MakeDir gid is correct} );
ok( sprintf( q{%lo}, $stat->mode() & 07777 ) eq sprintf( q{%lo},  0755),
    q{MakeDir mode is correct} );

ok(
    EFS::Util->MakeDir(
        path  => $testdir,
        user  => q{daemon},
        group => q{daemon},
        mode  => 02775,
    ),
    q{MakeDir success, daemon:deamon}
);

$stat = stat($testdir) || die qq{stat $ERRNO};

my ($daemon_uid) = (getpwnam(q{daemon}))[2];
my ($daemon_gid) = (getgrnam(q{daemon}))[2];

ok( $stat->uid() == $daemon_uid, q{MakeDir uid is correct} );
ok( $stat->gid() == $daemon_gid, q{MakeDir gid is correct} );
ok( sprintf( q{%lo}, $stat->mode() & 07777 ) eq sprintf( q{%lo},  02775),
    q{MakeDir mode is correct} );

# MakeDirHierarchy

throws_ok {
    EFS::Util->MakeDirHierarchy();
} qr{Mandatory parameter.* missing in call to}ms,
    q{MakeDirHierarchy parameter validation};

throws_ok {
    EFS::Util->MakeDirHierarchy(
        top    => q{/this/is/not/a/substring},
        bottom => q{/of/this},
        user   => q{root},
        group  => q{root},
        mode   => 0755,
    );
} qr{^Invalid directory hierachy specification\nTop (.*?) must be a substring of bottom (.*?)}ms,
    q{MakeDirHierarchy top/bottom sanity check};

ok(
    EFS::Util->MakeDirHierarchy(
        top    => qq{$testdir/a},
        bottom => qq{$testdir/a/b/c/d/e},
        user   => q{root},
        group  => q{root},
        mode   => 0755,
    ),
    q{MakeDirHierarchy success},
);

ok(
    EFS::Util->MakeDirHierarchy(
        top    => qq{$testdir///a/},
        bottom => qq{$testdir/a/b////c/d/e///f///},
        user   => q{root},
        group  => q{root},
        mode   => 0755,
    ),
    q{MakeDirHierarchy success, with funky slashes},
);

# RemoveDirHierarchy

throws_ok {
    EFS::Util->RemoveDirHierarchy();
} qr{^Mandatory parameter.* missing in call to}ms,
    q{RemoveDirHierarchy parameter validation};

throws_ok {
  EFS::Util->RemoveDirHierarchy(
      top   => qq{$testdir/a},
      empty => 1,
  );
} qr{Unable to rmtree $testdir/a: not empty}ms,
    q{RemoveDirHierarchy failure, --empty};

ok(
    EFS::Util->RemoveDirHierarchy(
        top   => qq{$testdir/a},
    ),
    q{RemoveDirHierarchy success},
);

ok( ! -d qq{$testdir/a}, qq{RemoveDirHierarchy directory is gone} );

# MakeLink

throws_ok {
    EFS::Util->MakeLink();
} qr{^Mandatory parameters '.*?', '.*?' missing in call to}ms,
    q{MakeLink parameter validation};

throws_ok {
    EFS::Util->MakeLink(
        from => $testdir,
        to   => q{/really/does/not/matter/for/this/test},
    );
} qr{^Unable to create or update symlink $testdir\nIt already exists as a directory}ms,
    q{MakeLink failure when from is not a symlink};

throws_ok {
    EFS::Util->MakeLink(
        from => q{/this/is/stupid},
        to   => q{/this/is/stupid},
    );
} qr{Invalid from and to specified: they match}ms,
    q{MakeLink stupid use case: from and to match};

ok(
    EFS::Util->MakeLink(
        from => qq{$testdir/link},
        to   => q{/where/it/points/does/not/matter},
    ),
    q{MakeLink basic success}
);

ok( -l qq{$testdir/link}, q{MakeLink exists} );
ok( readlink(qq{$testdir/link}) eq q{/where/it/points/does/not/matter},
    q{MakeLink readlink} );

ok(
    EFS::Util->MakeLink(
        from => qq{$testdir/link},
        to   => q{/some/other/bogus/path},
    ),
    q{MakeLink success when changing the link}
);

ok( -l qq{$testdir/link}, q{MakeLink exists} );
ok( readlink(qq{$testdir/link}) eq q{/some/other/bogus/path},
    q{MakeLink readlink} );

( my $funkylink = $testdir ) =~ s{/}{///}gms;

ok(
    EFS::Util->MakeLink(
        from => qq{$funkylink/link},
        to   => q{/one//more///time/with///funky/slashes////},
    ),
    q{MakeLink success when extra slashes are given}
);

ok( -l qq{$testdir/link}, q{MakeLink exists} );
ok( readlink(qq{$testdir/link}) eq q{/one/more/time/with/funky/slashes},
    q{MakeLink readlink} );

# RenameDir

mkdir( qq{$testdir/norename}, 0755 ) || die qq{mkdir: $ERRNO};

mkdir( qq{$testdir/rename}, 0755 ) || die qq{mkdir: $ERRNO};

throws_ok {
    EFS::Util->RenameDir();
} qr{^Mandatory parameter.* missing in call to}ms,
    q{RenameDir parameter validation};

throws_ok {
    EFS::Util->RenameDir(
        from => qq{$testdir/rename},
        to   => qq{$testdir/norename},
    );
} qr{Unable to rename .*/rename to .*/norename: target is a directory}ms,
    q{RenameDir failure when to is a directory};

symlink( qq{$testdir/norename}, qq{$testdir/renamelink} ) ||
    die qq{symlink: $ERRNO};

ok(
    EFS::Util->RenameDir(
        from => qq{$testdir/rename},
        to   => qq{$testdir/renamelink},
    ),
    q{RenameDir on top of symlink}
);

ok( ! -l qq{$testdir/renamelink}, q{RenameDir symlink is gone} );
ok( -d qq{$testdir/renamelink}, q{RenameDir directory} );
ok( ! -d qq{$testdir/rename}, q{RenameDir from is gone} );

# CreateLinkedDirectory

mkdir( qq{$testdir/linkto}, 0755 ) || die qq{mkdir: $ERRNO};
mkdir( qq{$testdir/linkto/bin}, 0755 ) || die qq{mkdir: $ERRNO};
mkdir( qq{$testdir/linkto/lib}, 0755 ) || die qq{mkdir: $ERRNO};
mkdir( qq{$testdir/linkfrom}, 0755 ) || die qq{mkdir: $ERRNO};

throws_ok {
    EFS::Util->CreateLinkedDirectory();
} qr{^Mandatory parameter.* missing in call to}ms,
    q{CreateLinkedDirectory parameter validation};

throws_ok {
    EFS::Util->CreateLinkedDirectory(
        from => qq{$testdir/nosuchfromdir},
        to   => qq{$testdir/nosuchtodir},
    );
} qr{Unable to opendir $testdir/nosuchtodir:}ms,
    q{CreateLinkedDirectory to dir missing};

throws_ok {
    EFS::Util->CreateLinkedDirectory(
        from => qq{$testdir/nosuchfromdir},
        to   => qq{$testdir/linkto},
    );
} qr{Unable to opendir $testdir/nosuchfromdir:}ms,
    q{CreateLinkedDirectory from dir missing};

mkdir( qq{$testdir/linkfrom/notalink}, 0755 ) || die qq{mkdir: $ERRNO};
symlink( q{foo}, qq{$testdir/linkfrom/boguslink} ) || die qq{symlink: $ERRNO};
symlink( q{wrongtarget}, qq{$testdir/linkfrom/bin} ) || die qq{symlink: $ERRNO};


throws_ok {
    EFS::Util->CreateLinkedDirectory(
        from => qq{$testdir/linkfrom},
        to   => qq{$testdir/linkto},
    );
} qr{Not a symlink: $testdir/linkfrom/notalink}ms,
    q{CreateLinkedDirectory from dir has a non-symlink};

rmdir( qq{$testdir/linkfrom/notalink} ) || die qq{rmdir: $ERRNO};

ok(
    EFS::Util->CreateLinkedDirectory(
        from => qq{$testdir/linkfrom},
        to   => qq{$testdir/linkto},
    ),
    q{CreateLinkedDirectory success},
);

ok( ! -l qq{$testdir/linkfrom/boguslink},
    q{CreateLinkedDirectory removed boguslink} );
ok( -l qq{$testdir/linkfrom/bin} &&
        readlink(qq{$testdir/linkfrom/bin}) eq qq{$testdir/linkto/bin},
    q{CreateLinkedDirectory fixed incorrect symlink} );
ok( -l qq{$testdir/linkfrom/lib} &&
        readlink(qq{$testdir/linkfrom/lib}) eq qq{$testdir/linkto/lib},
    q{CreateLinkedDirectory created missing symlink} );

# CloneDirectory

mkdir( qq{$testdir/clonefrom}, 0755 ) || die qq{mkdir: $ERRNO};
mkdir( qq{$testdir/clonefrom/bin}, 0755 ) || die qq{mkdir: $ERRNO};
mkdir( qq{$testdir/clonefrom/lib}, 0755 ) || die qq{mkdir: $ERRNO};

system( qq{echo $testdir/clonefrom/somepath > $testdir/clonefrom/bin/foo} );
$CHILD_ERROR and die qq{Unable to create bogus file\n};

throws_ok {
    EFS::Util->CloneDirectory();
} qr{^Mandatory parameter.* missing in call to}ms,
    q{CloneDirectory parameter validation};

throws_ok {
    EFS::Util->CloneDirectory(
        from => qq{$testdir/nosuchdir},
        to   => qq{$testdir/cloneto},
    );
} qr{No such directory: $testdir/nosuchdir}ms,
    q{CloneDirectory from dir missing};

ok(
    EFS::Util->CloneDirectory(
        from => qq{$testdir/clonefrom},
        to   => qq{$testdir/cloneto},
    ),
    q{CloneDirectory success, no replace},
);

ok( -d qq{$testdir/cloneto/lib}, q{CloneDirectory created lib dir} );
ok( -d qq{$testdir/cloneto/bin}, q{CloneDirectory created bin dir} );
ok( -f qq{$testdir/cloneto/bin/foo}, q{CloneDirectory created bin/foo file} );

system( qq{grep -q $testdir/clonefrom/somepath $testdir/cloneto/bin/foo} );
ok( $CHILD_ERROR == 0, q{CloneDirectory without replace work as expected} );

ok(
    EFS::Util->CloneDirectory(
        from    => qq{$testdir/clonefrom},
        to      => qq{$testdir/cloneto},
        replace => {
            qq{$testdir/clonefrom} => qq{$testdir/cloneto},
        },
    ),
    q{CloneDirectory success, with replace},
);

system( qq{grep -q $testdir/cloneto/somepath $testdir/cloneto/bin/foo} );
ok( $CHILD_ERROR == 0, q{CloneDirectory with replace works as expected} );

# That's all folks....

system(qq{rm -rf $testdir});
ok( ! -d $testdir, q{Testdir teardown});


