#!/usr/bin/env perl
use strict;
use warnings;
use English;

use blib;
use EFS;

use lib q{test/lib};
use EFS::Test;
use EFS::Test::Config;

use Fcntl qw{:mode};
use IO::Dir;
use Readonly;
use Test::More;
use Expect;
use File::Path qw(rmtree);
use File::Spec;
use File::Temp;
use File::Slurp;

my $test   = EFS::Test->new;
my $config = $test->get_config;

my $rdc_efsserver = $config->get_value( q{RemoteDevCell}, q{efsserver} );
my $sshcmd        = $config->get_value( q{Global},        q{sshcmd} );

my $ppid           = $PID;
my $username       = $test->get_primary_username;
my $groupname      = $test->get_primary_groupname;
my $home           = $test->get_primary_home;
my $command        = q{efs display metaproj bogusmeta};

my $expect_timeout = 10;
my $expect_delay   = 3; # Sleep inbetweemn calls to spawn/accept, or expect fails

##  set up a fake homedir, efsconfig auth file, and password,
##  because we mustn't change the user's environment for testing
my $fakehome       = File::Temp->newdir(
    q{efs_test_fakehome-XXXXXX}, TMPDIR => 1,
);
my $fake_efsconfig = File::Spec->catdir( $fakehome, qw{ .efsconfig } );
my $fakeauth       = File::Spec->catfile( $fake_efsconfig, qw{ auth } );

# If configured for GSSAPI, then we have to downgrade the server's
# authtype to PAM for the duration of this test.  Try to restore
# things in an END block.
if ( $config->get_value( q{Security}, q{kerberos_realm} ) ) {
    set_server_authtype( q{PAM} );
}

local $ENV{EFS_SERVERAUTH} = q{PAM};

$test->run_as_primary_user(
    q{Testing -authuser password},
    sub {

        local $ENV{EFS_CLI_PASSWORD} = q{correct};

        my $this = $test->efs_test_command(
            command =>
                q{efs create metaproj foo bar baz blah -authuser efstestuser},
            not     => 1,
            expect  => [
                q{The efs client is being run as efstestuser, therefore},
                q{absolutely no commands will be authorized.},
                q{Exit status from efs server not received.  Assuming failure...},
            ],
        );

        local $ENV{EFS_CLI_PASSWORD} = q{bogus};

        my $that = $test->efs_test_command(
            command =>
                q{efs create metaproj foo bar baz blah -authuser efstestuser},
            not     => 1,
            expect  => [
                q{Invalid authuser password (from EFS_CLI_PASSWORD)},
            ],
        );

        my $status = $this && $that;

        return $status;

    }
);


EFS::Util->MakeDir(
    path  => $fakehome->dirname,
    user  => q{root},
    group => q{root},
    mode  => 02775,
);
local $ENV{EFS_TEST_FAKE_PASSWORD} = q{totally_bogus_nonsense};


$test->run_as_primary_user(
    q{Verify home must be owned by the user},

    sub {

        local $ENV{HOME} = $fakehome;

        return $test->efs_test_command(
            command => $command,
            not     => 1,
            expect  => [
                qq{Directory $fakehome is not owned by $username},
                q{Access denied},
            ],
        );

    }
);

EFS::Util->MakeDir(
    path  => $fakehome->dirname,
    user  => $username,
    group => $groupname,
    mode  => 02775,
);

$test->run_as_primary_user(
    q{Verify home must have sane permissions},
    sub {

        local $ENV{HOME} = $fakehome;

        return $test->efs_test_command(
            command => $command,
            not     => 1,
            expect  => [
                qq{Directory $fakehome has invalid permissions},
                q{Access denied},
            ],
        );

    }
);

EFS::Util->MakeDir(
    path  => $fakehome->dirname,
    user  => $username,
    group => $groupname,
    mode  => 02755,
);

EFS::Util->MakeDir(
    path  => $fake_efsconfig,
    user  => q{root},
    group => q{root},
    mode  => 0755,
);

$test->run_as_primary_user(
    q{Verify .efsconfig must be owned by the user},
    sub {

        local $ENV{HOME} = $fakehome;

        return $test->efs_test_command(
            command => $command,
            not     => 1,
            expect  => [
                qq{Directory $fake_efsconfig is not owned by $username},
                q{Access denied},
            ],
        );

    }
);

EFS::Util->MakeDir(
    path  => $fake_efsconfig,
    user  => $username,
    group => $groupname,
    mode  => 0755,
);

$test->run_as_primary_user(
    q{Verify .efsconfig must have sane permissions},
    sub {

        local $ENV{HOME} = $fakehome;

        return $test->efs_test_command(
            command => $command,
            not     => 1,
            expect  => [
                qq{Directory $fake_efsconfig has invalid permissions},
                q{Access denied},
            ],
        );

    }
);

EFS::Util->MakeDir(
    path  => $fake_efsconfig,
    user  => $username,
    group => $groupname,
    mode  => 0700,
);

my $authfile = $fakeauth;

$test->run_as_primary_user(
    q{Verify auth file must have valid syntax},
    sub {

        local $ENV{HOME} = $fakehome;

        my $auth = IO::File->new($authfile, q{w+}) ||
            die qq{Unable to create $authfile: $ERRNO\n};

        $auth->print(qq{This is not an INI file format, now is it?\n});

        $auth->close;

        my $this = $test->efs_test_command(
            command => $command,
            not     => 1,
            expect  => [
                qq{Unable to parse $authfile},
                q{Access denied},
            ],
        );

        $auth = IO::File->new($authfile, q{w+}) ||
            die qq{Unable to create $authfile: $ERRNO\n};

        $auth->print(<<"END_OF_AUTHFILE");
[security]
password=$ENV{EFS_TEST_FAKE_PASSWORD}
END_OF_AUTHFILE

        $auth->close;
        my $that = $test->efs_test_command(
            command => $command,
            not     => 1,
            expect  => [
                q{Access denied},
            ],
        );

        my $status = $this && $that;

        return $status;

    }
);

my $efsconfig = File::Spec->catdir( $home, qw{ .efsconfig } );
$authfile     = File::Spec->catfile( $efsconfig, qw{ auth } );
delete $ENV{EFS_TEST_FAKE_PASSWORD};

SKIP: {

    if ( qx{uname} =~ m{FreeBSD}msx ) {
        skip q{Expect doesn't work right on FreeBSD}, 4;
    }

    Readonly my $SKIP_TEST_COUNT => 4;

    system qq{su $username -c 'chmod 755 $efsconfig'};
    $CHILD_ERROR and die qq{Unable to chmod $efsconfig\n};

    system qq{su $username -c 'chmod 644 $authfile'};
    $CHILD_ERROR and die qq{Unable to chmod $authfile\n};

    my $authcfg = Config::Tiny->read( $authfile );

    system qq{su $username -c 'chmod 600 $authfile'};
    $CHILD_ERROR and die qq{Unable to chmod $authfile\n};

    system qq{su $username -c 'chmod 700 $efsconfig'};
    $CHILD_ERROR and die qq{Unable to chmod $efsconfig\n};

    if ( not $authcfg ) {
        skip q{Unable to read primary user's authfile}
            . qq{ ('$authfile'): $ERRNO\n}, $SKIP_TEST_COUNT;
    }

    if ( not defined $authcfg->{security}->{password}
            or not length $authcfg->{security}->{password} ) {
        skip q{Password entry null or not found in primary user's authfile}
            . qq{ ('$authfile'): $ERRNO\n}, $SKIP_TEST_COUNT;
    }

    my $password = EFS::CLI->rotate13( $authcfg->{security}->{password} );

    #
    # This next test is VERY unstable.  It is commented, rather being
    # made a TODO, because you can not use a TODO block here.  The
    # run_as_primary_user method would have to be modified to support
    # the use of TODO, which is a lot of work for just this workaround...
    #

#     $test->run_as_primary_user(
#         q{Verify password caching works},
#         sub {
#             local $ENV{HOME} = $fakehome;

#             my $warning_seen   = 0;
#             my $expected_error = 0;

#             my $expect = Expect->new;

#             $expect->log_stdout(0);
#             $expect->spawn($command);

#             sleep $expect_delay; # Expect sucks....

#             $expect->expect(
#                 $expect_timeout,
#                 [
#                     qr{Password\s+for\s+$username:\s+}msx => sub {
#                         shift->send(qq{$password\n});
#                         exp_continue;
#                     }
#                 ],
#                 [
#                     qr{WARNING:\s+Password\s+cached\s+in\s+$fakeauth\s+with\s+light\s+encryption}msx => sub {
#                         $warning_seen = 1;
#                         exp_continue;
#                     }
#                 ],
#                 [
#                     qr{No \s+ such \s+ metaproj \s+ 'bogusmeta'}msx => sub {
#                         $expected_error = 1;
#                     }
#                 ],
#             );

#             if ( not $warning_seen ) {
#                 die qq{Warning message about password caching not seen\n};
#             }

#             if ( not $expected_error ) {
#                 my $error = $expect->before . $expect->after;
#                 $error =~ s{$password}{XXXXXXXX}gmsx;
#                 die(
#                     qq{Unable to test password submission:\n},
#                     $error,
#                 );
#             }

#             my $fakeconfig = Config::Tiny->read( $fakeauth ) ||
#                 die qq{Unable to open $fakeauth: $ERRNO\n};

#             my $cached_password =
#                 EFS::CLI->rotate13( $fakeconfig->{security}->{password} );

#             if ( $password ne $cached_password ) {
#                 die qq{Cached password does not match real password\n};
#             }

#             return 1;

#         }
#     );

    $test->run_as_primary_user(
        q{Verify disabling password caching works via config file},
        sub {

            local $ENV{HOME} = $fakehome;

            write_file( $fakeauth, <<"END_OF_AUTHFILE" );
[security]
no-auth-cache=true
END_OF_AUTHFILE

            my $prompt_seen = 0;
            my $warning_seen = 0;
            my $expected_error = 0;

            my $expect = Expect->new;

            $expect->log_file( q{/var/tmp/20_cli_pam_08.out} );
            $expect->log_stdout(0);
            $expect->spawn($command);

            sleep $expect_delay; # Expect sucks....

            $expect->expect(
                $expect_timeout,
                [
                    qr{Password\s+for\s+$username:\s+}msx => sub {
                        shift->send(qq{$password\n});
                        $prompt_seen = 1;
                        exp_continue;
                    }
                ],
                [
                    qr{WARNING:\s+Password\s+cached\s+in\s+$fakeauth\s+with\s+light\s+encryption}msx => sub {
                        $warning_seen = 1;
                        exp_continue;
                    }
                ],
                [
                    qr{No \s+ such \s+ metaproj \s+ 'bogusmeta'}msx => sub {
                        $expected_error = 1;
                    }
                ],
            );

            if ( not $prompt_seen ) {
                die qq{Never saw the password prompt\n};
            }

            if ( $warning_seen ) {
                die qq{Saw the warning indicating the password was cached\n};
            }

            if ( not $expected_error ) {
                my $error = $expect->before . $expect->after;
                $error =~ s{$password}{XXXXXXXX}gmsx;
                die(
                    qq{Unable to test password submission:\n},
                    $error,
                );
            }

            my $fakeconfig = Config::Tiny->read( $fakeauth ) ||
                die qq{Unable to open $fakeauth: $ERRNO\n};

            if ( $fakeconfig->{security}->{password} ) {
                die qq{Password was incorrectly cached\n};
            }

            return 1;

        }
    );

    $test->run_as_primary_user(
        q{Verify disabling password caching works via CLI option},
        sub {

            local $ENV{HOME} = $fakehome;

            unlink($fakeauth) ||
                die qq{Unable to unlink $fakeauth: $ERRNO\n};

            my $prompt_seen = 0;
            my $warning_seen = 0;
            my $expected_error = 0;

            my $expect = Expect->new;

            $expect->log_file( q{/var/tmp/20_cli_pam_09.out} );
            $expect->log_stdout(0);
            $expect->spawn(qq{$command --no-auth-cache});

            sleep $expect_delay;

            $expect->expect(
                $expect_timeout,
                [
                    qr{Password\s+for\s+$username:\s+}msx => sub {
                        shift->send(qq{$password\n});
                        $prompt_seen = 1;
                        exp_continue;
                    }
                ],
                [
                    qr{WARNING:\s+Password\s+cached\s+in\s+$fakeauth\s+with\s+light\s+encryption}msx => sub {
                        $warning_seen = 1;
                        exp_continue;
                    }
                ],
                [
                    qr{No \s+ such \s+ metaproj \s+ 'bogusmeta'}msx => sub {
                        $expected_error = 1;
                    }
                ],
            );

            if ( not $prompt_seen ) {
                die qq{Never saw the password prompt\n};
            }

            if ( $warning_seen ) {
                die qq{Saw the warning indicating the password was cached\n};
            }

            if ( not $expected_error ) {
                my $error = $expect->before . $expect->after;
                $error =~ s{$password}{XXXXXXXX}gmsx;
                die(
                    qq{Unable to test password submission:\n},
                    $error,
                );
            }

            if ( -f $fakeauth ) {
                die qq{Password was incorrectly cached\n};
            }

            return 1;

        }
    );

    $test->run_as_primary_user(
        q{Verify disabling password caching works via env var},
        sub {

            my $authcfg = Config::Tiny->read( $authfile ) ||
                    die qq{Unable to open primary user's authfile: $ERRNO\n};

            my $password = EFS::CLI->rotate13( $authcfg->{security}->{password} );

            local $ENV{HOME} = $fakehome;

            local $ENV{EFS_NO_AUTH_CACHE} = 1;

            my $expected_error = 0;

            my $expect = Expect->new;

            $expect->log_file( q{/var/tmp/20_cli_pam_10.out} );
            $expect->log_stdout(0);
            $expect->spawn($command);

            sleep $expect_delay;

            $expect->expect(
                $expect_timeout,
                [
                    qr{Password\s+for\s+$username:\s+}msx => sub {
                        shift->send(qq{$password\n});
                        exp_continue;
                    }
                ],
                [
                    qr{No \s+ such \s+ metaproj \s+ 'bogusmeta'}msx => sub {
                        $expected_error = 1;
                    }
                ],
            );

            if ( not $expected_error ) {
                my $error = $expect->before . $expect->after;
                $error =~ s{$password}{XXXXXXXX}gmsx;
                die(
                    qq{Unable to test password submission:\n},
                    $error,
                );
            }

            if ( -f $fakeauth ) {
                die qq{Password was incorrectly cached\n};
            }

            return 1;

        }
    );

} #SKIP


END {
    if ( $ppid == $PID ) {
        done_testing();
        if ( $config->get_value( q{Security}, q{kerberos_realm} ) ) {
            set_server_authtype( q{GSSAPI} );
        }
    }
}

sub set_server_authtype {

    my $authtype = shift;

    my $serverconf = read_file( q{/etc/efs/server.conf} );
    $serverconf =~ s{authtype=.*?$}{authtype=$authtype}gmsx;
    write_file( q{/etc/efs/server.conf}, $serverconf );

    system( qq{rsync --rsh='$sshcmd' -avz }
                . qq{/etc/efs/server.conf $rdc_efsserver:/etc/efs/server.conf} );
    $CHILD_ERROR and die qq{Unable to rsync /etc/efs/server.conf to $rdc_efsserver};

    return 1;

}

1;

