#!/usr/bin/env perl

use strict;
use warnings;
use English;

use Test::More tests => 56;

use lib q{test/lib};
use EFS::Test;

my $test = EFS::Test->new;

my $username  = $test->get_primary_username;
my $groupname = $test->get_primary_groupname;

my @setup = (
    q{efs display metaproj testmeta},
    q{efs display project testmeta testproj},
    q{efs create release testmeta testproj testrel},
    q{efs create install testmeta testproj testrel common},
);

foreach my $setup (@setup) {
    $test->efs_test_command(
        name    => qq{(setup) $setup},
        command => $setup,
        bail    => 1,
    );
}

foreach my $action ( qw( execlinks audit ) ) {

    $test->efs_test_command(
        run_as  => q{efsops},
        command => qq{efs $action metaproj testmeta},
    );

    $test->efs_test_command(
        command => qq{efs history release testmeta testproj testrel -command '$action metaproj' -count 1},
        filter => qr{^ \s+ \d+}msx,
        expect => qr{ $action \s+ release \s+ testmeta \s+ testproj \s+ testrel }msx,
    );

    my @history = ();

    $test->efs_test_command(
        command => qq{efs history any -command '$action metaproj' -count 3},
        filter  => qr{^ \s+ \d+}msx,
        output  => \@history,
        expect  => [
            qr{ $action \s+ metaproj \s+ testmeta }msx,
            qr{ $action \s+ project \s+ testmeta \s+ testproj }msx,
            qr{ $action \s+ release \s+ testmeta \s+ testproj \s+ testrel }msx,
        ],
    );

    verify_nested_logfiles( @history );

}

$test->efs_test_command(
    command => q{efs chown metaproj testmeta},
);

my @history = ();

$test->efs_test_command(
    command => q{efs history any -command 'chown metaproj' -count 3},
    filter  => qr{^ \s+ \d+}msx,
    output  => \@history,
    expect  => [
        qr{ chown \s+ metaproj \s+ testmeta }msx,
        qr{ chown \s+ project \s+ testmeta \s+ testproj }msx,
        qr{ chown \s+ release \s+ testmeta \s+ testproj \s+ testrel }msx,
    ],
);

verify_nested_logfiles( @history );

$test->efs_test_command(
    command => q{efs history release testmeta testproj testrel -command 'chown metaproj' -count 1},
    filter  => qr{^ \s+ \d+}msx,
    expect  => qr{ chown \s+ release \s+ testmeta \s+ testproj \s+ testrel }msx,
);

$test->efs_test_command(
    command => q{efs setstage release testmeta testproj testrel rejected},
);

$test->efs_test_command(
    command => q{efs destroy release testmeta testproj testrel},
);

$test->efs_test_command(
    run_as  => q{efsops},
    command => q{efs purge any},
);

$test->efs_test_command(
    command => q{efs history release testmeta testproj testrel -command 'purge any' -count 1},
    filter  => qr{^ \s+ \d+}msx,
    expect  => qr{ purge \s+ release \s+ testmeta \s+ testproj \s+ testrel }msx,
);

@history = ();

$test->efs_test_command(
    command => q{efs history any -command 'purge any' -count 4},
    filter  => qr{^ \s+ \d+}msx,
    output  => \@history,
    expect  => [
        qr{ purge \s+ any }msx,
        qr{ purge \s+ metaproj \s+ testmeta }msx,
        qr{ purge \s+ project \s+ testmeta \s+ testproj }msx,
        qr{ purge \s+ release \s+ testmeta \s+ testproj \s+ testrel }msx,
    ],
);

verify_nested_logfiles( @history );

foreach my $userspec ( q{bogususername}, qq{bogususername,$username} ) {
    $test->efs_test_command(
        command => qq{efs history any -byuser $userspec},
        not     => 1,
        expect  => q{No such username 'bogususername'},
    );
}

foreach my $groupspec ( q{bogusgroupname}, qq{bogusgroupname,$groupname} ) {
    $test->efs_test_command(
        command => qq{efs history any -bygroup $groupspec},
        not     => 1,
        expect  => q{No such groupname 'bogusgroupname'},
    );
}

$test->efs_test_command(
    command => qq{efs history any -byuser $username -count 1},
    filter  => qr{^ \s+ \d+}msx,
    expect  => qr{ destroy \s+ release \s+ testmeta \s+ testproj \s+ testrel }msx,
);

$test->efs_test_command(
    command => qq{efs history any -bygroup $groupname -count 1},
    filter  => qr{^ \s+ \d+}msx,
    expect  => qr{ destroy \s+ release \s+ testmeta \s+ testproj \s+ testrel }msx,
);

my @teardown = (
    q{efs purge release testmeta testproj testrel -now},
);

foreach my $teardown (@teardown) {
    $test->efs_test_command(
        run_as  => q{efsops},
        name    => qq{(teardown) $teardown},
        command => $teardown,
    );
}

exit 0;

sub verify_nested_logfiles {

    my @history = @_;

    my @ids = ();

    foreach my $history ( reverse @history ) {
        if ( my ($id) = ($history =~ m{^ \s+ (\d+)}msx) ) {
            push @ids, $id;
        }
    }

    my %logfiles = ();
    my %count = ();

    foreach my $id ( @ids ) {

        my @logfile = ();

        $test->efs_test_command(
            command => qq{efs history logfile $id},
            output  => \@logfile,
        );

        foreach my $line ( @logfile ) {
            $line =~ s{^ \[ [^\]]+ \] \s+ }{}msx;
            $line =~ s{ \s+ $ }{}msx;
            next if $line =~ m{ ^ _EFS_ }msx;
            $logfiles{$id} .= qq{$line\n};
            $count{$id}++;
        }

    }

    foreach my $index ( 0 .. ($#ids - 1) ) {
        my $id = $ids[$index];
        my $parent = $ids[$index+1];
        ok( $logfiles{$parent} =~ m{\Q$logfiles{$id}\E}msx,
            qq{Logfile for $id is contained inside logfile for $parent} );
        ok( $count{$parent} > $count{$id},
            qq{Logfile for $id is shorter than logfile for $parent} );
    }

    return;

}
