#!/usr/bin/env perl

use strict;
use warnings;
use English;

use Test::More tests => 42;

use blib;
use EFS;

use lib q{test/lib};
use EFS::Test;

## no critic: RequireExtendedFormatting

my $test   = EFS::Test->new;
my $config = $test->get_config;

my $username  = $test->get_primary_username;
my $groupname = $test->get_primary_groupname;

my $clone_mpr  = q{testmeta testclone clone};
my $clone_path = q{testmeta/testclone/clone};

my $child_mpr  = q{testmeta testclone child};
my $child_path = q{testmeta/testclone/child};

my @setup_one = (
    [ 0, q{efs display metaproj testmeta} ],
    [ 1, q{efs display project testmeta testclone} ],
    [ 0, qq{efs create project testmeta testclone} ],
    [ 0, qq{efs create release $clone_mpr} ],
    [ 0, qq{efs create install $clone_mpr common} ],
    [ 0, qq{touch /efs/dev/$clone_path/install/common/olddata} ],
    [ 0, qq{touch /efs/dev/$clone_path/src/oldsrc} ],
    [ 0, qq{touch /efs/dev/$clone_path/build/oldbuild} ],
);

foreach my $setup ( @setup_one ) {
    my ($not,$command) = @{ $setup };
    $test->efs_test_command(
        name    => qq{(setup, phase one) $command},
        not     => $not,
        command => $command,
        bail    => 1,
    );
}

$test->efs_test_command(
    command => q{efs clone release testmeta testclone child clone},
    not     => 1,
    expect  => qq{Clone $clone_path is not locked},
);

my @setup_two = (
    qq{efs checkpoint release $clone_mpr},
    qq{efs dist release $clone_mpr --global},
);

foreach my $setup ( @setup_two ) {
    $test->efs_test_command(
        name    => qq{(setup, phase two) $setup},
        command => $setup,
    );
}

$test->efs_test_command(
    command => qq{efs clone release testmeta testclone toolong clone},
    not     => 1,
    expect  => [
        qq{Cloned releases must have names of matching length},
        qq{length(clone) != length(toolong)},
    ],
);

$test->efs_test_command(
    command => qq{efs clone release $child_mpr clone},
    filter  => qr{Cloned|Unlocking}ms,
    expect  => [
        qq{Cloned directory /efs/dev/$clone_path/install/common to /efs/dev/$child_path/install/common},
        qq{Unlocking install /efs/dev/$child_path/install/common},
        qq{Cloned install $child_path/common from $clone_path/common},
        qq{Cloned release $clone_path to $child_path},
    ],
);

$test->efs_test_file(
    file  => qq{/efs/dev/$child_path/install/common/olddata},
    user  => $username,
    group => $groupname,
);

$test->efs_test_command(
    command => qq{efs display release $clone_mpr},
    expect  => [
        qr{Metaproj Project Release Stage Locked}ms,
        qr{=+}ms,
        qr{testmeta testclone clone dev yes}ms,
        q{The following releases are clones of this one:},
        qq{$child_path},
        qr{Metaproj Project Release Install}ms,
        qr{=+}ms,
        qr{testmeta testclone clone common}ms,
    ],
);

$test->efs_test_command(
    command => qq{efs display release $child_mpr},
    expect  => [
        qr{Metaproj Project Release Stage Locked}ms,
        qr{=+}ms,
        qr{testmeta testclone child dev no}ms,
        qq{This release is a clone of $clone_path},
        qr{Metaproj Project Release Install}ms,
        qr{=+}ms,
        qr{testmeta testclone child common}ms,
    ],
);

$test->efs_test_command(
    command => qq{efs unclone release $child_mpr},
    expect  => qq{Release $child_path is no longer a clone},
);

$test->efs_test_command(
    command => qq{efs display release $clone_mpr},
    expect  => [
        qr{Metaproj Project Release Stage Locked}ms,
        qr{=+}ms,
        qr{testmeta testclone clone dev yes}ms,
        qr{Metaproj Project Release Install}ms,
        qr{=+}ms,
        qr{testmeta testclone clone common}ms,
    ],
);

purge_child_release();

$test->efs_test_command(
    command => qq{efs clone release $child_mpr clone --src},
    filter  => qr{Cloned|Unlocking}ms,
    expect  => [
        qq{Cloned directory /efs/dev/$clone_path/install/common to /efs/dev/$child_path/install/common},
        qq{Unlocking install /efs/dev/$child_path/install/common},
        qq{Cloned install $child_path/common from $clone_path/common},
        qq{Cloned directory /efs/dev/$clone_path/src to /efs/dev/$child_path/src},
        qq{Unlocking directory /efs/dev/$child_path/src},
        qq{Unlocking directory /efs/dev/$child_path/build},
        qq{Cloned release $clone_path to $child_path},
    ],
);

purge_child_release();

$test->efs_test_command(
    command => qq{efs clone release $child_mpr clone --build},
    filter  => qr{Cloned|Unlocking}ms,
    expect  => [
        qq{Cloned directory /efs/dev/$clone_path/install/common to /efs/dev/$child_path/install/common},
        qq{Unlocking install /efs/dev/$child_path/install/common},
        qq{Cloned install $child_path/common from $clone_path/common},
        qq{Cloned directory /efs/dev/$clone_path/build to /efs/dev/$child_path/build},
        qq{Unlocking directory /efs/dev/$child_path/src},
        qq{Unlocking directory /efs/dev/$child_path/build},
        qq{Cloned release $clone_path to $child_path},            
    ],
);

my @teardown_one = ();

foreach my $teardown ( @teardown_one ) {
    my $run_as = $teardown =~ m{--now}ms ? q{efsops} : q{primary_user};
    $test->efs_test_command(
        run_as  => $run_as,
        name    => qq{(teardown, phase one) $teardown},
        command => $teardown,
    );
}

$test->efs_test_command(
    command => qq{touch /efs/dev/$child_path/install/common/newdata},
);

$test->efs_test_command(
    command => qq{efs checkpoint release $child_mpr},
);

$test->efs_test_command(
    command => qq{efs dist release $child_mpr --global},
);

$test->efs_test_command(
    command => qq{efs setstage release $child_mpr prod},
    expect  => [
        qq{Validating the checkpoint state for release '$child_path'.},
        qq{Validating the audit state for release '$child_path'.},
        qq{Changed the stage of $child_path from dev to prod},
        qq{Release $child_path is a clone of $clone_path, but},
        qq{$clone_path has a stage of 'dev'},
        q{Cloning optimizations will not be available when},
        qq{$child_path is disted to prod cells},
    ],
);

$test->efs_test_command(
    command => qq{efs setstage release $clone_mpr prod},
);

$test->efs_test_command(
    command => qq{efs dist release $child_mpr --celltype prod --global},
    not     => 1,
    expect  => [
        qq{Clone $clone_path has not been disted to all cells.},
        q{To dist without clone optimizations, use --noclone},
    ],
);

$test->efs_test_command(
    command => qq{efs dist release $child_mpr --celltype prod --noclone --global},
    filter  => qr{\bolddata\b}ms,
    expect  => qr{olddata}ms,
);

$test->efs_test_command(
    command => qq{efs display release $child_mpr},
    expect  => [
        qr{Metaproj Project Release Stage Locked}ms,
        qr{=+}ms,
        qr{testmeta testclone child prod yes}ms,
        qr{Metaproj Project Release Install}ms,
        qr{=+}ms,
        qr{testmeta testclone child common}ms,
    ],
);

$test->efs_test_command(
    command => q{efs history project testmeta testclone --count 23},
    expect  => [
        qr{Status Authuser Request}ms,
        qr{-+}ms,
        qr{1 $username create project testmeta testclone}ms,
        qr{1 $username create release $clone_mpr}ms,
        qr{1 $username create install $clone_mpr common}ms,
        qr{0 $username clone release $child_mpr clone}ms,
        qr{1 $username checkpoint release $clone_mpr}ms,
        qr{1 $username dist release $clone_mpr}ms,
        qr{0 $username clone release testmeta testclone toolong clone}ms,
        qr{1 $username clone release $child_mpr clone}ms,
        qr{1 $username unclone release $child_mpr}ms,
        qr{1 $username setstage release $child_mpr rejected}ms,
        qr{1 $username destroy release $child_mpr}ms,
        qr{1 $username purge release $child_mpr --now}ms,
        qr{1 $username clone release $child_mpr clone --src}ms,
        qr{1 $username setstage release $child_mpr rejected}ms,
        qr{1 $username destroy release $child_mpr}ms,
        qr{1 $username purge release $child_mpr --now}ms,
        qr{1 $username clone release $child_mpr clone --build}ms,
        qr{1 $username checkpoint release $child_mpr}ms,
        qr{1 $username dist release $child_mpr}ms,
        qr{1 $username setstage release $child_mpr prod}ms,
        qr{1 $username setstage release $clone_mpr prod}ms,
        qr{0 $username dist release $child_mpr --celltype prod}ms,
        qr{1 $username dist release $child_mpr --celltype prod --noclone}ms,
    ],
);

my @teardown = (
    qq{efs setstage release $child_mpr deprecated},
    qq{efs destroy release $child_mpr},
    qq{efs purge release $child_mpr --now},
    qq{efs setstage release $clone_mpr deprecated},
    qq{efs destroy release $clone_mpr},
    qq{efs purge release $clone_mpr --now},
);

push @teardown, q{efs destroy project testmeta testclone};

foreach my $teardown ( @teardown ) {
    my $run_as = $teardown =~ m{--now}ms ? q{efsops} : q{primary_user};
    $test->efs_test_command(
        run_as  => $run_as,
        name    => qq{(teardown) $teardown},
        command => $teardown,
    );
}

exit 0;

sub purge_child_release {

    my @commands = (
        qq{efs setstage release $child_mpr rejected},
        qq{efs destroy release $child_mpr},
        qq{efs purge release $child_mpr --now},
    );

    foreach my $command ( @commands ) {
        $test->efs_test_command( command => $command );
    }

}
